# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any

import numpy as np
from numpy.typing import NDArray

LOG = logging.getLogger(__name__)


class ViewCacheArray:
    """A class that provides a caching mechanism for writing to a NumPy-like array.

    The is initialized with a NumPy-like array, a shape and a list to reindex the first
    dimension. The array is used to store the final data, while the cache is used to
    temporarily store the data before flushing it to the array.

    The `flush` method copies the contents of the cache to the final array.
    """

    def __init__(self, array: NDArray[Any], *, shape: tuple[int, ...], indexes: list[int]):
        """Initialize the ViewCacheArray.

        Parameters
        ----------
        array : NDArray[Any]
            The NumPy-like array to store the final data.
        shape : tuple[int, ...]
            The shape of the cache array.
        indexes : list[int]
            List to reindex the first dimension.
        """
        assert len(indexes) == shape[0], (len(indexes), shape[0])
        self.array = array
        self.dtype = array.dtype
        self.cache = np.full(shape, np.nan, dtype=self.dtype)
        self.indexes = indexes

    def __setitem__(self, key: tuple[int, ...], value: NDArray[Any]) -> None:
        """Set the value in the cache array at the specified key.

        Parameters
        ----------
        key : tuple[int, ...]
            The index key to set the value.
        value : NDArray[Any]
            The value to set in the cache array.
        """
        self.cache[key] = value

    def flush(self) -> None:
        """Copy the contents of the cache to the final array."""
        for i in range(self.cache.shape[0]):
            global_i = self.indexes[i]
            self.array[global_i] = self.cache[i]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from functools import wraps
from typing import Any
from typing import Callable
from typing import List
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import NDArray

from .dataset import FullIndex
from .dataset import Shape
from .dataset import TupleIndex


def _tuple_with_slices(t: TupleIndex, shape: Shape) -> Tuple[TupleIndex, Tuple[int, ...]]:
    """Replace all integers in a tuple with slices, so we preserve the dimensionality.

    Parameters:
    t (TupleIndex): The tuple index to process.
    shape (Shape): The shape of the array.

    Returns:
    Tuple[TupleIndex, Tuple[int, ...]]: A tuple containing the modified index and the changes.
    """
    result = tuple(slice(i, i + 1) if isinstance(i, int) else i for i in t)
    changes = tuple(j for (j, i) in enumerate(t) if isinstance(i, int))
    result = tuple(slice(*s.indices(shape[i])) for (i, s) in enumerate(result))

    return result, changes


def _extend_shape(index: TupleIndex, shape: Shape) -> TupleIndex:
    """Extend the shape of the index to match the shape of the array.

    Parameters:
    index (TupleIndex): The index to extend.
    shape (Shape): The shape of the array.

    Returns:
    TupleIndex: The extended index.
    """
    if Ellipsis in index:
        if index.count(Ellipsis) > 1:
            raise IndexError("Only one Ellipsis is allowed")
        ellipsis_index = index.index(Ellipsis)
        index = list(index)
        index[ellipsis_index] = slice(None)
        while len(index) < len(shape):
            index.insert(ellipsis_index, slice(None))
        index = tuple(index)

    while len(index) < len(shape):
        index = index + (slice(None),)

    return index


def _index_to_tuple(index: FullIndex, shape: Shape) -> TupleIndex:
    """Convert an index to a tuple index.

    Parameters:
    index (FullIndex): The index to convert.
    shape (Shape): The shape of the array.

    Returns:
    TupleIndex: The converted tuple index.
    """
    if isinstance(index, int):
        return _extend_shape((index,), shape)
    if isinstance(index, slice):
        return _extend_shape((index,), shape)
    if isinstance(index, tuple):
        return _extend_shape(index, shape)
    if index is Ellipsis:
        return _extend_shape((Ellipsis,), shape)
    raise ValueError(f"Invalid index: {index}")


def index_to_slices(index: Union[int, slice, Tuple], shape: Shape) -> Tuple[TupleIndex, Tuple[int, ...]]:
    """Convert an index to a tuple of slices, with the same dimensionality as the shape.

    Parameters:
    index (Union[int, slice, Tuple]): The index to convert.
    shape (Shape): The shape of the array.

    Returns:
    Tuple[TupleIndex, Tuple[int, ...]]: A tuple containing the slices and the changes.
    """
    return _tuple_with_slices(_index_to_tuple(index, shape), shape)


def apply_index_to_slices_changes(result: NDArray[Any], changes: Tuple[int, ...]) -> NDArray[Any]:
    """Apply changes to the result array based on the slices.

    Parameters:
    result (NDArray[Any]): The result array.
    changes (Tuple[int, ...]): The changes to apply.

    Returns:
    NDArray[Any]: The modified result array.
    """
    if changes:
        shape = result.shape
        for i in changes:
            assert shape[i] == 1, (i, changes, shape)
        result = np.squeeze(result, axis=changes)
    return result


def update_tuple(t: Tuple, index: int, value: Any) -> Tuple[Tuple, Any]:
    """Replace the elements of a tuple at the given index with a new value.

    Parameters:
    tp (Tuple): The original tuple.
    index (int): The index to update.
    value (Any): The new value.

    Returns:
    Tuple[Tuple, Any]: The updated tuple and the previous value.
    """
    t = list(t)
    prev = t[index]
    t[index] = value
    return tuple(t), prev


def length_to_slices(index: slice, lengths: List[int]) -> List[Union[slice, None]]:
    """Convert an index to a list of slices, given the lengths of the dimensions.

    Parameters:
    index (slice): The index to convert.
    lengths (List[int]): The lengths of the dimensions.

    Returns:
    List[Union[slice, None]]: A list of slices.
    """
    total = sum(lengths)
    start, stop, step = index.indices(total)

    result = []

    pos = 0
    for length in lengths:
        end = pos + length

        b = max(pos, start)
        e = min(end, stop)

        p = None
        if b <= e:
            if (b - start) % step != 0:
                b = b + step - (b - start) % step
            b -= pos
            e -= pos

            if 0 <= b < e:
                p = slice(b, e, step)

        result.append(p)

        pos = end

    return result


def _as_tuples(index: Tuple) -> Tuple:
    """Convert elements of the index to tuples if they are lists or arrays.

    Parameters:
    index (Tuple): The index to convert.

    Returns:
    Tuple: The converted index.
    """

    def _(i: Any) -> Any:
        if hasattr(i, "tolist"):
            # NumPy arrays, TensorFlow tensors, etc.
            i = i.tolist()
            assert not isinstance(i[0], bool), "Mask not supported"
            return tuple(i)

        if isinstance(i, list):
            return tuple(i)

        return i

    return tuple(_(i) for i in index)


def expand_list_indexing(method: Callable[..., NDArray[Any]]) -> Callable[..., NDArray[Any]]:
    """Allows to use slices, lists, and tuples to select data from the dataset.
    Zarr does not support indexing with lists/arrays directly,
    so we need to implement it ourselves.

    Parameters:
    method (Callable[..., NDArray[Any]]): The method to wrap.

    Returns:
    Callable[..., NDArray[Any]]: The wrapped method.
    """

    @wraps(method)
    def wrapper(self: Any, index: FullIndex) -> NDArray[Any]:
        if not isinstance(index, tuple):
            return method(self, index)

        if not any(isinstance(i, (list, tuple)) for i in index):
            return method(self, index)

        which: List[int] = []
        for i, idx in enumerate(index):
            if isinstance(idx, (list, tuple)):
                which.append(i)

        assert which, "No list index found"

        if len(which) > 1:
            raise IndexError("Only one list index is allowed")

        which = which[0]
        index = _as_tuples(index)
        result = []
        for i in index[which]:
            index, _ = update_tuple(index, which, slice(i, i + 1))
            result.append(method(self, index))

        return np.concatenate(result, axis=which)

    return wrapper


def make_slice_or_index_from_list_or_tuple(indices: List[int]) -> Union[List[int], slice]:
    """Convert a list or tuple of indices to a slice or an index, if possible.

    Parameters:
    indices (List[int]): The list or tuple of indices.

    Returns:
    Union[List[int], slice]: The slice or index.
    """
    if len(indices) < 2:
        return indices

    step = indices[1] - indices[0]

    if step > 0 and all(indices[i] - indices[i - 1] == step for i in range(1, len(indices))):
        return slice(indices[0], indices[-1] + step, step)

    return indices



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

import numpy as np
from anemoi.utils.dates import frequency_to_timedelta
from numpy.typing import NDArray

from .dataset import Dataset
from .dataset import FullIndex
from .dataset import Shape
from .dataset import TupleIndex
from .debug import Node
from .debug import debug_indexing
from .forwards import Forwards
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import update_tuple

LOG = logging.getLogger(__name__)


class InterpolateFrequency(Forwards):
    """A class to represent a dataset with interpolated frequency."""

    def __init__(self, dataset: Dataset, frequency: str) -> None:
        """Initialize the InterpolateFrequency class.

        Parameters
        ----------
        dataset : Dataset
            The dataset to be interpolated.
        frequency : str
            The interpolation frequency.
        """
        super().__init__(dataset)
        self._frequency = frequency_to_timedelta(frequency)

        self.seconds = self._frequency.total_seconds()
        other_seconds = dataset.frequency.total_seconds()

        self.seconds = int(self.seconds)
        assert self.seconds == self._frequency.total_seconds()

        other_seconds = int(other_seconds)
        assert other_seconds == dataset.frequency.total_seconds()

        if self.seconds >= other_seconds:
            raise ValueError(
                f"Interpolate frequency {self._frequency} must be more frequent than dataset frequency {dataset.frequency}"
            )

        if other_seconds % self.seconds != 0:
            raise ValueError(
                f"Interpolate frequency {self._frequency}  must be a multiple of the dataset frequency {dataset.frequency}"
            )

        self.ratio = other_seconds // self.seconds
        self.alphas = np.linspace(0, 1, self.ratio + 1)
        self.other_len = len(dataset)

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index: TupleIndex) -> NDArray[Any]:
        """Get the interpolated data for a tuple index.

        Parameters
        ----------
        index : TupleIndex
            The tuple index to retrieve data from.

        Returns
        -------
        NDArray[Any]
            The interpolated data for the tuple index.
        """
        index, changes = index_to_slices(index, self.shape)
        index, previous = update_tuple(index, 0, slice(None))
        result = self._get_slice(previous)
        return apply_index_to_slices_changes(result[index], changes)

    def _get_slice(self, s: slice) -> NDArray[Any]:
        """Get the interpolated data for a slice.

        Parameters
        ----------
        s : slice
            The slice to retrieve data from.

        Returns
        -------
        NDArray[Any]
            The interpolated data for the slice.
        """
        return np.stack([self[i] for i in range(*s.indices(self._len))])

    @debug_indexing
    def __getitem__(self, n: FullIndex) -> NDArray[Any]:
        """Get the interpolated data at the specified index.

        Parameters
        ----------
        n : FullIndex
            The index to retrieve data from.

        Returns
        -------
        NDArray[Any]
            The interpolated data at the specified index.
        """
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        if n < 0:
            n += self._len

        if n == self._len - 1:
            # Special case for the last element
            return self.forward[-1]

        i = n // self.ratio
        x = n % self.ratio

        if x == 0:
            # No interpolation needed
            return self.forward[i]

        alpha = self.alphas[x]

        assert 0 < alpha < 1, alpha
        return self.forward[i] * (1 - alpha) + self.forward[i + 1] * alpha

    def __len__(self) -> int:
        """Get the length of the interpolated dataset.

        Returns
        -------
        int
            The length of the interpolated dataset.
        """
        return (self.other_len - 1) * self.ratio + 1

    @property
    def frequency(self) -> datetime.timedelta:
        """Get the interpolation frequency."""
        return self._frequency

    @cached_property
    def dates(self) -> NDArray[np.datetime64]:
        """Get the interpolated dates."""
        result = []
        deltas = [np.timedelta64(self.seconds * i, "s") for i in range(self.ratio)]
        for d in self.forward.dates[:-1]:
            for i in deltas:
                result.append(d + i)
        result.append(self.forward.dates[-1])
        return np.array(result)

    @property
    def shape(self) -> Shape:
        """Get the shape of the interpolated dataset."""
        return (self._len,) + self.forward.shape[1:]

    def tree(self) -> Node:
        """Get the tree representation of the dataset.

        Returns
        -------
        Node
            The tree representation of the dataset.
        """
        return Node(self, [self.forward.tree()], frequency=self.frequency)

    @cached_property
    def missing(self) -> Set[int]:
        """Get the missing data indices."""
        result = []
        j = 0
        for i in range(self.other_len):
            missing = i in self.forward.missing
            for _ in range(self.ratio):
                if missing:
                    result.append(j)
                j += 1

        result = set(x for x in result if x < self._len)
        return result

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Get the metadata specific to the InterpolateFrequency subclass.

        Returns
        -------
        Dict[str, Any]
            The metadata specific to the InterpolateFrequency subclass.
        """
        return {
            # "frequency": frequency_to_string(self._frequency),
        }


class InterpolateNearest(Forwards):
    def __init__(
        self, dataset: Dataset, interpolate_variables: List[str], max_distance: Optional[float] = None
    ) -> None:
        """Initialize the InterpolateNearest class.

        Parameters
        ----------
        dataset : Dataset
            The dataset to be interpolated.
        interpolate_variables : List[str]
            The variables to be interpolated.
        max_distance : Optional[float], optional
            The maximum distance for nearest neighbor search, by default None.
        """
        from ..grids import nearest_grid_points

        super().__init__(dataset)
        self.vars = interpolate_variables
        self.var_mask = {
            v: np.where(~np.isnan(dataset[0, dataset.name_to_index[v]].squeeze()))[0] for v in interpolate_variables
        }
        self.dataset = dataset
        self.ngp_var = {
            v: nearest_grid_points(
                dataset.latitudes[self.var_mask[v]],
                dataset.longitudes[self.var_mask[v]],
                dataset.latitudes,
                dataset.longitudes,
                max_distance=max_distance,
            )
            for v in interpolate_variables
        }

    def tree(self) -> Node:
        return Node(self, [self.forward.tree()], interpolate_variables=self.vars)

    @property
    def shape(self) -> Shape:
        return self.forward.shape

    @property
    def metadata(self) -> Dict[str, Any]:
        return self.forward.metadata()

    @staticmethod
    def slice_len(slice_obj: slice) -> int:
        # Compute the length of the slice
        return max(0, (slice_obj.stop - slice_obj.start + (slice_obj.step or 1) - 1) // (slice_obj.step or 1))

    @expand_list_indexing
    def _get_tuple(self, index: TupleIndex) -> NDArray[Any]:
        index, changes = index_to_slices(index, self.shape)
        source_data = self.forward[index[0]]
        target_data = source_data.copy()
        vars_to_interpolate = [self.forward.name_to_index[v] for v in self.vars]
        for v, i in zip(self.vars, vars_to_interpolate):
            target_data[:, i, ...] = source_data[:, i][..., self.var_mask[v][self.ngp_var[v]]]
        result = target_data[(slice(None),) + index[1:]]
        return apply_index_to_slices_changes(result, changes)

    def __getitem__(self, index: Union[int, slice, Tuple[Union[int, slice], ...]]) -> NDArray[Any]:
        if isinstance(index, (int, slice)):
            index = (index, slice(None), slice(None), slice(None))
        return self._get_tuple(index)

    def subclass_metadata_specific(self) -> Dict[str, Any]:
        return {
            "interpolate_variables": self.vars,
        }



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import os
import textwrap
from functools import wraps
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import List
from typing import Optional

from anemoi.utils.text import Tree
from numpy.typing import NDArray

if TYPE_CHECKING:
    from .dataset import Dataset

LOG = logging.getLogger(__name__)

DEBUG_ZARR_LOADING = int(os.environ.get("DEBUG_ZARR_LOADING", "0"))
DEBUG_ZARR_INDEXING = int(os.environ.get("DEBUG_ZARR_INDEXING", "0"))

DEPTH = 0

# TODO: make numpy arrays read-only
# a.flags.writeable = False


def css(name: str) -> str:
    """Get the CSS content from a file.

    Parameters
    ----------
    name : str
        The name of the CSS file.

    Returns
    -------
    str
        The CSS content.
    """
    path = os.path.join(os.path.dirname(__file__), f"{name}.css")
    with open(path) as f:
        return f"<style>{f.read()}</style>"


class Node:
    """A class to represent a node in a dataset tree."""

    def __init__(self, dataset: "Dataset", kids: List[Any], **kwargs: Any) -> None:
        """Initializes a Node object.

        Parameters
        ----------
        dataset : Dataset
            The dataset associated with the node.
        kids : List[Any]
            List of child nodes.
        kwargs : Any
            Additional keyword arguments.
        """
        self.dataset = dataset
        self.kids = kids
        self.kwargs = kwargs

    def _put(self, indent: int, result: List[str]) -> None:
        """Helper method to add the node representation to the result list.

        Parameters
        ----------
        indent : int
            Indentation level.
        result : List[str]
            List to store the node representation.
        """

        def _spaces(indent: int) -> str:
            return " " * indent if indent else ""

        result.append(f"{_spaces(indent)}{self.dataset.__class__.__name__}")
        for k, v in self.kwargs.items():
            if isinstance(v, (list, tuple)):
                v = ", ".join(str(i) for i in v)
                v = textwrap.shorten(v, width=40, placeholder="...")
            result.append(f"{_spaces(indent+2)}{k}: {v}")
        for kid in self.kids:
            kid._put(indent + 2, result)

    def __repr__(self) -> str:
        """Returns the string representation of the node.

        Returns
        -------
        str
            String representation of the node.
        """
        result: List[str] = []
        self._put(0, result)
        return "\n".join(result)

    def graph(self, digraph: List[str], nodes: dict) -> None:
        """Generates a graph representation of the node.

        Parameters
        ----------
        digraph : List[str]
            List to store the graph representation.
        nodes : dict
            Dictionary to store the node labels.
        """
        label = self.dataset.label  # dataset.__class__.__name__.lower()
        if self.kwargs:
            param = []
            for k, v in self.kwargs.items():
                if k == "path" and isinstance(v, str):
                    v = os.path.basename(v)
                if isinstance(v, (list, tuple)):
                    v = ", ".join(str(i) for i in v)
                else:
                    v = str(v)
                v = textwrap.shorten(v, width=40, placeholder="...")
                # if len(self.kwargs) == 1:
                #     param.append(v)
                # else:
                param.append(f"{k}={v}")
            label = f'{label}({",".join(param)})'

        label += "\n" + "\n".join(
            textwrap.shorten(str(v), width=40, placeholder="...")
            for v in (
                self.dataset.dates[0],
                self.dataset.dates[-1],
                self.dataset.frequency,
                self.dataset.shape,
                self.dataset.variables,
            )
        )

        nodes[f"N{id(self)}"] = label
        for kid in self.kids:
            digraph.append(f"N{id(self)} -> N{id(kid)}")
            kid.graph(digraph, nodes)

    def digraph(self) -> str:
        """Returns the graph representation of the node.

        Returns
        -------
        str
            Graph representation of the node.
        """
        digraph = ["digraph {"]
        digraph.append("node [shape=box];")
        nodes = {}

        self.graph(digraph, nodes)

        for node, label in nodes.items():
            digraph.append(f'{node} [label="{label}"];')

        digraph.append("}")
        return "\n".join(digraph)

    def _html(self, indent: str, rows: List[List[str]]) -> None:
        """Helper method to add the node representation to the HTML rows.

        Parameters
        ----------
        indent : str
            Indentation level.
        rows : List[List[str]]
            List to store the HTML rows.
        """
        kwargs = {}

        for k, v in self.kwargs.items():
            if k == "path" and isinstance(v, str):
                v = v[::-1]
            if isinstance(v, (list, tuple)):
                v = ", ".join(str(i) for i in v)
            else:
                v = str(v)
            v = textwrap.shorten(v, width=80, placeholder="...")
            if k == "path":
                v = v[::-1]
            kwargs[k] = v
        label = self.dataset.label
        label = f'<span class="dataset">{label}</span>'
        if len(kwargs) == 1:
            k, v = list(kwargs.items())[0]
            rows.append([indent] + [label, v])
        else:
            rows.append([indent] + [label])

            for k, v in kwargs.items():
                rows.append([indent] + [f"<span class='param'>{k}</span>", f"<span class='param'>{v}</span>"])

        for kid in self.kids:
            kid._html(indent + "&nbsp;&nbsp;&nbsp;", rows)

    def html(self) -> str:
        """Returns the HTML representation of the node.

        Returns
        -------
        str
            HTML representation of the node.
        """
        result = [css("debug")]

        result.append('<table class="dataset">')
        rows = []

        self._html("", rows)

        for r in rows:
            s = " ".join(str(x) for x in r)
            result.append(f"<tr><td>{s}</td></tr>")

        result.append("</table>")
        return "\n".join(result)

    def _as_tree(self, tree: Any) -> None:
        """Helper method to add the node representation to the tree.

        Parameters
        ----------
        tree : Any
            Tree object to store the node representation.
        """
        for kid in self.kids:
            n = tree.node(kid)
            kid._as_tree(n)

    def as_tree(self) -> Tree:
        """Returns the tree representation of the node.

        Returns
        -------
        Tree
            Tree representation of the node.
        """

        tree = Tree(self)
        self._as_tree(tree)
        return tree

    @property
    def summary(self) -> str:
        """Returns the summary of the node."""
        return self.dataset.label

    def as_dict(self) -> dict:
        """Returns the dictionary representation of the node.

        Returns
        -------
        dict
            Dictionary representation of the node.
        """
        return {}


class Source:
    """A class used to follow the provenance of a data point."""

    def __init__(self, dataset: Any, index: int, source: Optional[Any] = None, info: Optional[Any] = None) -> None:
        """Initializes a Source object.

        Parameters
        ----------
        dataset : Any
            The dataset associated with the source.
        index : int
            Index of the data point.
        source : Optional[Any], optional
            Source of the data point, by default None.
        info : Optional[Any], optional
            Additional information, by default None.
        """
        self.dataset = dataset

        self.index = index
        self.source = source
        self.info = info

    def __repr__(self) -> str:
        """Returns the string representation of the source.

        Returns
        -------
        str
            String representation of the source.
        """
        p = s = self.source
        while s is not None:
            p = s
            s = s.source

        return f"{self.dataset}[{self.index}, {self.dataset.variables[self.index]}] ({p})"

    def target(self) -> Any:
        """Returns the target source.

        Returns
        -------
        Any
            Target source.
        """
        p = s = self.source
        while s is not None:
            p = s
            s = s.source
        return p

    def dump(self, depth: int = 0) -> None:
        """Dumps the source information.

        Parameters
        ----------
        depth : int, optional
            Indentation level, by default 0.
        """
        print(" " * depth, self)
        if self.source is not None:
            self.source.dump(depth + 1)


def _debug_indexing(method: Callable[..., NDArray[Any]]) -> Callable[..., NDArray[Any]]:
    """Decorator to debug indexing methods.

    Parameters
    ----------
    method : Callable[..., NDArray[Any]]
        The method to be decorated.

    Returns
    -------
    Callable[..., NDArray[Any]]
        The decorated method.
    """

    @wraps(method)
    def wrapper(self: Any, index: Any) -> NDArray[Any]:
        global DEPTH
        # if isinstance(index, tuple):
        print("  " * DEPTH, "->", self, method.__name__, index)
        DEPTH += 1
        result = method(self, index)
        DEPTH -= 1
        # if isinstance(index, tuple):
        print("  " * DEPTH, "<-", self, method.__name__, result.shape)
        return result

    return wrapper


def _identity(method: Callable[..., NDArray[Any]]) -> Callable[..., NDArray[Any]]:
    """Identity function.

    Parameters
    ----------
    method : Callable[..., NDArray[Any]]
        Input method.

    Returns
    -------
    Callable[..., NDArray[Any]]
        The input method.
    """
    return method


if DEBUG_ZARR_INDEXING:
    debug_indexing = _debug_indexing
else:
    debug_indexing = _identity


def debug_zarr_loading(on_off: int) -> None:
    """Enables or disables Zarr loading debugging.

    Parameters
    ----------
    on_off : int
        1 to enable debugging, 0 to disable.
    """
    global DEBUG_ZARR_LOADING
    DEBUG_ZARR_LOADING = on_off



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from collections import defaultdict
from typing import Any
from typing import List

import earthkit.data as ekd
import numpy as np
from anemoi.transform.fields import new_field_from_numpy
from anemoi.transform.fields import new_fieldlist_from_list
from earthkit.meteo.wind.array import polar_to_xy

from .legacy import legacy_filter


@legacy_filter(__file__)
def execute(
    context: Any,
    input: List[Any],
    wind_speed: str,
    wind_dir: str,
    u_component: str = "u",
    v_component: str = "v",
    in_radians: bool = False,
) -> ekd.FieldList:
    """Convert wind speed and direction to u and v components.

    Parameters
    ----------
    context : Any
        The context for the execution.
    input : List[Any]
        The input data fields.
    wind_speed : str
        The name of the wind speed parameter.
    wind_dir : str
        The name of the wind direction parameter.
    u_component : str, optional
        The name for the u component. Defaults to "u".
    v_component : str, optional
        The name for the v component. Defaults to "v".
    in_radians : bool, optional
        Whether the wind direction is in radians. Defaults to False.

    Returns
    -------
    ekd.FieldList
        The resulting field array with u and v components.
    """

    result = []

    wind_params = (wind_speed, wind_dir)
    wind_pairs = defaultdict(dict)

    for f in input:
        key = f.metadata(namespace="mars")
        param = key.pop("param")

        if param not in wind_params:
            result.append(f)
            continue

        key = tuple(key.items())

        if param in wind_pairs[key]:
            raise ValueError(f"Duplicate wind component {param} for {key}")

        wind_pairs[key][param] = f

    for _, pairs in wind_pairs.items():
        if len(pairs) != 2:
            raise ValueError("Missing wind component")

        magnitude = pairs[wind_speed]
        direction = pairs[wind_dir]

        # assert speed.grid_mapping == dir.grid_mapping
        if in_radians:
            direction = np.rad2deg(direction)

        u, v = polar_to_xy(magnitude.to_numpy(flatten=True), direction.to_numpy(flatten=True))

        result.append(new_field_from_numpy(magnitude, u, param=u_component))
        result.append(new_field_from_numpy(direction, v, param=v_component))

    return new_fieldlist_from_list(result)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from typing import TYPE_CHECKING
from typing import Any
from typing import Set

# from .dataset import FullIndex
# from .dataset import Shape
# from .dataset import TupleIndex
from .misc import _open_dataset
from .misc import _save_dataset
from .misc import add_dataset_path
from .misc import add_named_dataset

if TYPE_CHECKING:
    from .dataset import Dataset

LOG = logging.getLogger(__name__)

__all__ = [
    "open_dataset",
    "MissingDateError",
    "add_dataset_path",
    "add_named_dataset",
]


class MissingDateError(Exception):
    pass


def _convert(x: Any) -> Any:
    """Convert OmegaConf objects to standard Python containers.

    Parameters
    ----------
    x : Any
        The object to convert.

    Returns
    -------
    Any
        The converted object.
    """
    if isinstance(x, list):
        return [_convert(a) for a in x]

    if isinstance(x, tuple):
        return tuple(_convert(a) for a in x)

    if isinstance(x, dict):
        return {k: _convert(v) for k, v in x.items()}

    if x.__class__.__name__ in ("DictConfig", "ListConfig"):
        from omegaconf import OmegaConf

        return OmegaConf.to_container(x, resolve=True)

    return x


def open_dataset(*args: Any, **kwargs: Any) -> "Dataset":
    """Open a dataset.

    Parameters
    ----------
    *args : Any
        Positional arguments.
    **kwargs : Any
        Keyword arguments.

    Returns
    -------
    Dataset
        The opened dataset.
    """
    # That will get rid of OmegaConf objects

    args, kwargs = _convert(args), _convert(kwargs)

    ds = _open_dataset(*args, **kwargs)
    ds = ds.mutate()
    ds.arguments = {"args": args, "kwargs": kwargs}
    ds._check()
    return ds


def save_dataset(recipe: dict, zarr_path: str, n_workers: int = 1) -> None:
    """Open a dataset and save it to disk.

    Parameters
    ----------
    recipe : dict
        Recipe used with open_dataset (not a dataset creation recipe).
    zarr_path : str
        Path to store the obtained anemoi dataset to disk.
    n_workers : int
        Number of workers to use for parallel processing. If none, sequential processing will be performed.
    """
    _save_dataset(recipe, zarr_path, n_workers)


def list_dataset_names(*args: Any, **kwargs: Any) -> list[str]:
    """List the names of datasets.

    Parameters
    ----------
    *args : Any
        Positional arguments.
    **kwargs : Any
        Keyword arguments.

    Returns
    -------
    list of str
        The list of dataset names.
    """
    ds = _open_dataset(*args, **kwargs)
    names: Set[str] = set()
    ds.get_dataset_names(names)
    return sorted(names)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from typing import Any

from anemoi.utils.registry import Registry

LOG = logging.getLogger(__name__)


source_registry = Registry(__name__)


def create_source(context: Any, config: Any) -> Any:
    """Create a source based on the provided configuration.

    Parameters
    ----------
    context : Any
        The context in which the source is created.
    config : Any
        The configuration for the source.

    Returns
    -------
    Any
        The created source.
    """
    return source_registry.from_config(config, context)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple

from .dataset import Dataset
from .dataset import FullIndex
from .debug import Node
from .forwards import Combined
from .misc import _auto_adjust
from .misc import _open

LOG = logging.getLogger(__name__)


class ZipBase(Combined):
    """Base class for handling zipped datasets."""

    def __init__(self, datasets: List[Any], check_compatibility: bool = True) -> None:
        """Initialize ZipBase with a list of datasets.

        Parameters
        ----------
        datasets : List[Any]
            List of datasets.
        check_compatibility : bool, optional
            Flag to check compatibility of datasets, by default True.
        """
        self._check_compatibility = check_compatibility
        super().__init__(datasets)

    def swap_with_parent(self, parent: Any) -> Any:
        """Swap datasets with the parent.

        Parameters
        ----------
        parent : Any
            Parent dataset.

        Returns
        -------
        Any
            New parent dataset with swapped datasets.
        """
        new_parents = [parent.clone(ds) for ds in self.datasets]
        return self.clone(new_parents)

    def clone(self, datasets: List[Any]) -> "ZipBase":
        """Clone the ZipBase with new datasets.

        Parameters
        ----------
        datasets : List[Any]
            List of new datasets.

        Returns
        -------
        ZipBase
            Cloned ZipBase instance.
        """
        return self.__class__(datasets, check_compatibility=self._check_compatibility)

    def tree(self) -> Node:
        """Get the tree representation of the datasets.

        Returns
        -------
        Node
            Tree representation of the datasets.
        """
        return Node(self, [d.tree() for d in self.datasets], check_compatibility=self._check_compatibility)

    def __len__(self) -> int:
        """Get the length of the smallest dataset.

        Returns
        -------
        int
            Length of the smallest dataset.
        """
        return min(len(d) for d in self.datasets)

    def __getitem__(self, n: FullIndex) -> Tuple[Any, ...]:
        """Get the item at the specified index from all datasets.

        Parameters
        ----------
        n : FullIndex
            Index to retrieve.

        Returns
        -------
        Tuple[Any, ...]
            Tuple of items from all datasets.
        """
        return tuple(d[n] for d in self.datasets)

    def check_same_resolution(self, d1: Dataset, d2: Dataset) -> None:
        """Check if two datasets have the same resolution.

        Parameters
        ----------
        d1 : Dataset
            First dataset.
        d2 : Dataset
            Second dataset.
        """
        pass

    def check_same_grid(self, d1: Dataset, d2: Dataset) -> None:
        """Check if two datasets have the same grid.

        Parameters
        ----------
        d1 : Dataset
            First dataset.
        d2 : Dataset
            Second dataset.
        """
        pass

    def check_same_variables(self, d1: Dataset, d2: Dataset) -> None:
        """Check if two datasets have the same variables.

        Parameters
        ----------
        d1 : Dataset
            First dataset.
        d2 : Dataset
            Second dataset.
        """
        pass

    @cached_property
    def missing(self) -> Set[int]:
        """Get the set of missing indices from all datasets."""
        result: Set[int] = set()
        for d in self.datasets:
            result = result | d.missing
        return result

    @property
    def shape(self) -> Tuple[Any, ...]:
        """Get the shape of all datasets."""
        return tuple(d.shape for d in self.datasets)

    @property
    def field_shape(self) -> Tuple[Any, ...]:
        """Get the field shape of all datasets."""
        return tuple(d.shape for d in self.datasets)

    @property
    def latitudes(self) -> Tuple[Any, ...]:
        """Get the latitudes of all datasets."""
        return tuple(d.latitudes for d in self.datasets)

    @property
    def longitudes(self) -> Tuple[Any, ...]:
        """Get the longitudes of all datasets."""
        return tuple(d.longitudes for d in self.datasets)

    @property
    def dtype(self) -> Tuple[Any, ...]:
        """Get the data types of all datasets."""
        return tuple(d.dtype for d in self.datasets)

    @property
    def grids(self) -> Tuple[Any, ...]:
        """Get the grids of all datasets."""
        return tuple(d.grids for d in self.datasets)

    @property
    def statistics(self) -> Tuple[Any, ...]:
        """Get the statistics of all datasets."""
        return tuple(d.statistics for d in self.datasets)

    @property
    def resolution(self) -> Tuple[Any, ...]:
        """Get the resolution of all datasets."""
        return tuple(d.resolution for d in self.datasets)

    @property
    def name_to_index(self) -> Tuple[Any, ...]:
        """Get the name to index mapping of all datasets."""
        return tuple(d.name_to_index for d in self.datasets)

    def check_compatibility(self, d1: Dataset, d2: Dataset) -> None:
        """Check compatibility between two datasets.

        Parameters
        ----------
        d1 : Dataset
            First dataset.
        d2 : Dataset
            Second dataset.
        """
        if self._check_compatibility:
            super().check_compatibility(d1, d2)


class Zip(ZipBase):
    """Class for handling zipped datasets."""

    pass


class XY(ZipBase):
    """Class for handling XY datasets."""

    pass


def xy_factory(args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> XY:
    """Factory function to create an XY instance.

    Parameters
    ----------
    args : Tuple[Any, ...]
        Positional arguments.
    kwargs : Dict[str, Any]
        Keyword arguments.

    Returns
    -------
    XY
        An instance of XY.
    """
    if "xy" in kwargs:
        xy = kwargs.pop("xy")
    else:
        xy = [kwargs.pop("x"), kwargs.pop("y")]

    assert len(args) == 0
    assert isinstance(xy, (list, tuple))

    datasets = [_open(e) for e in xy]
    datasets, kwargs = _auto_adjust(datasets, kwargs)

    assert len(datasets) == 2

    check_compatibility = kwargs.pop("check_compatibility", True)

    return XY(datasets, check_compatibility=check_compatibility)._subset(**kwargs)


def zip_factory(args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> Zip:
    """Factory function to create a Zip instance.

    Parameters
    ----------
    args : Tuple[Any, ...]
        Positional arguments.
    kwargs : Dict[str, Any]
        Keyword arguments.

    Returns
    -------
    Zip
        An instance of Zip.
    """
    zip = kwargs.pop("zip")
    assert len(args) == 0
    assert isinstance(zip, (list, tuple))

    datasets = [_open(e) for e in zip]
    datasets, kwargs = _auto_adjust(datasets, kwargs)

    check_compatibility = kwargs.pop("check_compatibility", True)

    return Zip(datasets, check_compatibility=check_compatibility)._subset(**kwargs)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
import logging
import time
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from typing import Any

import tqdm
from anemoi.utils.humanize import seconds_to_human

from . import Command

LOG = logging.getLogger(__name__)


def task(what: str, options: dict, *args: Any, **kwargs: Any) -> Any:
    """Make sure `import Creator` is done in the sub-processes, and not in the main one.

    Parameters
    ----------
    what : str
        The task to be executed.
    options : dict
        Options for the task.
    *args : Any
        Additional arguments.
    **kwargs : Any
        Additional keyword arguments.

    Returns
    -------
    Any
        The result of the task.
    """
    now = datetime.datetime.now()
    LOG.info(f"🎬 Task {what}({args},{kwargs}) starting")

    from anemoi.datasets.create import creator_factory

    options = {k: v for k, v in options.items() if v is not None}

    c = creator_factory(what.replace("-", "_"), **options)
    result = c.run()

    LOG.info(f"🏁 Task {what}({args},{kwargs}) completed ({datetime.datetime.now()-now})")
    return result


class Create(Command):
    """Create a dataset."""

    internal = True
    timestamp = True

    def add_arguments(self, command_parser: Any) -> None:
        """Add command line arguments to the parser.

        Parameters
        ----------
        command_parser : Any
            The command line argument parser.
        """
        command_parser.add_argument(
            "--overwrite",
            action="store_true",
            help="Overwrite existing files. This will delete the target dataset if it already exists.",
        )
        command_parser.add_argument(
            "--test",
            action="store_true",
            help="Build a small dataset, using only the first dates. And, when possible, using low resolution and less ensemble members.",
        )
        command_parser.add_argument("config", help="Configuration yaml file defining the recipe to create the dataset.")
        command_parser.add_argument("path", help="Path to store the created data.")
        group = command_parser.add_mutually_exclusive_group()
        group.add_argument("--threads", help="Use `n` parallel thread workers.", type=int, default=0)
        group.add_argument("--processes", help="Use `n` parallel process workers.", type=int, default=0)
        command_parser.add_argument("--trace", action="store_true")

    def run(self, args: Any) -> None:
        """Execute the create command.

        Parameters
        ----------
        args : Any
            Command line arguments.
        """
        now = time.time()
        if args.threads + args.processes:
            self.parallel_create(args)
        else:
            self.serial_create(args)
        LOG.info(f"Create completed in {seconds_to_human(time.time()-now)}")

    def serial_create(self, args: Any) -> None:
        """Create the dataset in serial mode.

        Parameters
        ----------
        args : Any
            Command line arguments.
        """
        options = vars(args)
        options.pop("command")
        options.pop("threads")
        options.pop("processes")

        task("init", options)
        task("load", options)
        task("finalise", options)

        task("init_additions", options)
        task("run_additions", options)
        task("finalise_additions", options)

        task("patch", options)

        task("cleanup", options)
        task("verify", options)

    def parallel_create(self, args: Any) -> None:
        """Create the dataset in parallel mode.

        Parameters
        ----------
        args : Any
            Command line arguments.
        """
        """Some modules, like fsspec do not work well with fork()
        Other modules may not be thread safe. So we implement
        parallel loadining using multiprocessing before any
        of the modules are imported.
        """

        options = vars(args)
        options.pop("command")

        threads = options.pop("threads")
        processes = options.pop("processes")

        use_threads = threads > 0
        options["use_threads"] = use_threads

        if use_threads:
            ExecutorClass = ThreadPoolExecutor
        else:
            ExecutorClass = ProcessPoolExecutor

        with ExecutorClass(max_workers=1) as executor:
            total = executor.submit(task, "init", options).result()

        futures = []

        parallel = threads + processes
        with ExecutorClass(max_workers=parallel) as executor:
            for n in range(total):
                opt = options.copy()
                opt["parts"] = f"{n+1}/{total}"
                futures.append(executor.submit(task, "load", opt))

            for future in tqdm.tqdm(
                as_completed(futures), desc="Loading", total=len(futures), colour="green", position=parallel + 1
            ):
                future.result()

        with ExecutorClass(max_workers=1) as executor:
            executor.submit(task, "finalise", options).result()

        with ExecutorClass(max_workers=1) as executor:
            executor.submit(task, "init-additions", options).result()

        with ExecutorClass(max_workers=parallel) as executor:
            opt = options.copy()
            opt["parts"] = f"{n+1}/{total}"
            futures.append(executor.submit(task, "load", opt))
            for n in range(total):
                futures.append(executor.submit(task, "load-additions", opt))

            for future in tqdm.tqdm(
                as_completed(futures),
                desc="Computing additions",
                total=len(futures),
                colour="green",
                position=parallel + 1,
            ):
                future.result()

        with ExecutorClass(max_workers=1) as executor:
            executor.submit(task, "finalise-additions", options).result()
            executor.submit(task, "patch", options).result()
            executor.submit(task, "cleanup", options).result()
            executor.submit(task, "verify", options).result()


command = Create



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any
from typing import Dict
from typing import Tuple

import numpy as np
from numpy.typing import NDArray

from .dataset import Dataset
from .dataset import FullIndex
from .dataset import Shape
from .debug import Node
from .forwards import Forwards
from .forwards import GivenAxis
from .indexing import apply_index_to_slices_changes
from .indexing import index_to_slices
from .indexing import update_tuple
from .misc import _auto_adjust
from .misc import _open

LOG = logging.getLogger(__name__)

OFFSETS = dict(number=1, numbers=1, member=0, members=0)


class Number(Forwards):
    """A class to represent a subset of ensemble members from a dataset."""

    def __init__(self, forward: Dataset, **kwargs: Any) -> None:
        """Initializes a Number object.

        Parameters
        ----------
        forward : Any
            The dataset to forward.
        kwargs : Any
            Additional keyword arguments specifying the members.
        """
        super().__init__(forward)

        self.members = []
        for key, values in kwargs.items():
            if not isinstance(values, (list, tuple)):
                values = [values]
            self.members.extend([int(v) - OFFSETS[key] for v in values])

        self.members = sorted(set(self.members))
        for n in self.members:
            if not (0 <= n < forward.shape[2]):
                raise ValueError(f"Member {n} is out of range. `number(s)` is one-based, `member(s)` is zero-based.")

        self.mask = np.array([n in self.members for n in range(forward.shape[2])], dtype=bool)
        self._shape, _ = update_tuple(forward.shape, 2, len(self.members))

    @property
    def shape(self) -> Shape:
        """Returns the shape of the dataset."""
        return self._shape

    def __getitem__(self, index: FullIndex) -> NDArray[Any]:
        """Retrieves data from the dataset based on the given index.

        Parameters
        ----------
        index : FullIndex
            Index specifying the data to retrieve.

        Returns
        -------
        NDArray[Any]
            Data array from the dataset based on the index.
        """
        if isinstance(index, int):
            result = self.forward[index]
            result = result[:, self.mask, :]
            return result

        if isinstance(index, slice):
            result = self.forward[index]
            result = result[:, :, self.mask, :]
            return result

        index, changes = index_to_slices(index, self.forward.shape)
        result = self.forward[index]
        result = result[:, :, self.mask, :]
        return apply_index_to_slices_changes(result, changes)

    def tree(self) -> Node:
        """Generates a hierarchical tree structure for the dataset.

        Returns
        -------
        Node
            A Node object representing the dataset.
        """
        return Node(self, [self.forward.tree()], numbers=[n + 1 for n in self.members])

    def metadata_specific(self, **kwargs: Any) -> Dict[str, Any]:
        """Returns metadata specific to the Number object.

        Parameters
        ----------
        kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Dict[str, Any]
            Metadata specific to the Number object.
        """
        return {
            "numbers": [n + 1 for n in self.members],
        }

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Returns metadata specific to the Number object."""
        return {}


class Ensemble(GivenAxis):
    """A class to represent an ensemble of datasets combined along a given axis."""

    def tree(self) -> Node:
        """Generates a hierarchical tree structure for the ensemble datasets.

        Returns
        -------
        Node
            A Node object representing the ensemble datasets.
        """
        return Node(self, [d.tree() for d in self.datasets])

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Get the metadata specific to the forwards subclass.

        Returns:
        Dict[str, Any]: The metadata specific to the forwards subclass.
        """
        return {}


def ensemble_factory(args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> Ensemble:
    """Factory function to create an Ensemble object.

    Parameters
    ----------
    args : Tuple[Any, ...]
        Positional arguments.
    kwargs : dict
        Keyword arguments.

    Returns
    -------
    Ensemble
        An Ensemble object.
    """
    if "grids" in kwargs:
        raise NotImplementedError("Cannot use both 'ensemble' and 'grids'")

    ensemble = kwargs.pop("ensemble")
    axis = kwargs.pop("axis", 2)
    assert len(args) == 0
    assert isinstance(ensemble, (list, tuple))

    datasets = [_open(e) for e in ensemble]
    datasets, kwargs = _auto_adjust(datasets, kwargs)

    return Ensemble(datasets, axis=axis)._subset(**kwargs)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import textwrap
import threading
from functools import wraps
from typing import Any
from typing import Callable

LOG = logging.getLogger(__name__)


thread_local = threading.local()
TRACE = 0


def enable_trace(on_off: int) -> None:
    """Enables or disables tracing.

    Parameters
    ----------
    on_off : int
        1 to enable tracing, 0 to disable.
    """
    global TRACE
    TRACE = on_off


def step(action_path: list[str]) -> str:
    """Returns a formatted string representing the action path.

    Parameters
    ----------
    action_path : list of str
        The action path.

    Returns
    -------
    str
        The formatted action path.
    """
    return f"[{'.'.join(action_path)}]"


def trace(emoji: str, *args: Any) -> None:
    """Logs a trace message with an emoji.

    Parameters
    ----------
    emoji : str
        The emoji to use.
    *args : Any
        The arguments to log.
    """
    if not TRACE:
        return

    if not hasattr(thread_local, "TRACE_INDENT"):
        thread_local.TRACE_INDENT = 0

    print(emoji, " " * thread_local.TRACE_INDENT, *args)


def trace_datasource(method: Callable) -> Callable:
    """Decorator to trace the datasource method.

    Parameters
    ----------
    method : Callable
        The method to decorate.

    Returns
    -------
    Callable
        The wrapped method.
    """

    @wraps(method)
    def wrapper(self, *args: Any, **kwargs: Any) -> Any:

        if not hasattr(thread_local, "TRACE_INDENT"):
            thread_local.TRACE_INDENT = 0

        trace(
            "🌍",
            "=>",
            step(self.action_path),
            self._trace_datasource(*args, **kwargs),
        )
        thread_local.TRACE_INDENT += 1
        result = method(self, *args, **kwargs)
        thread_local.TRACE_INDENT -= 1
        trace(
            "🍎",
            "<=",
            step(self.action_path),
            textwrap.shorten(repr(result), 256),
        )
        return result

    return wrapper


def trace_select(method: Callable) -> Callable:
    """Decorator to trace the select method.

    Parameters
    ----------
    method : Callable
        The method to decorate.

    Returns
    -------
    Callable
        The wrapped method.
    """

    @wraps(method)
    def wrapper(self, *args: Any, **kwargs: Any) -> Any:
        if not hasattr(thread_local, "TRACE_INDENT"):
            thread_local.TRACE_INDENT = 0
        trace(
            "👓",
            "=>",
            ".".join(self.action_path),
            self._trace_select(*args, **kwargs),
        )
        thread_local.TRACE_INDENT += 1
        result = method(self, *args, **kwargs)
        thread_local.TRACE_INDENT -= 1
        trace(
            "🍍",
            "<=",
            ".".join(self.action_path),
            textwrap.shorten(repr(result), 256),
        )
        return result

    return wrapper



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
import warnings
from abc import abstractmethod
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set

import numpy as np
from numpy.typing import NDArray

from .dataset import Dataset
from .dataset import FullIndex
from .dataset import Shape
from .dataset import TupleIndex
from .debug import debug_indexing
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import length_to_slices
from .indexing import update_tuple

LOG = logging.getLogger(__name__)


class Forwards(Dataset):
    """A class to represent a dataset that forwards its properties and methods to another dataset."""

    def __init__(self, forward: Dataset) -> None:
        """Initializes a Forwards object.

        Parameters
        ----------
        forward : Dataset
            The forward dataset.
        """
        self.forward = forward.mutate()

    def __len__(self) -> int:
        """Returns the length of the forward dataset.

        Returns
        -------
        int
            Length of the forward dataset.
        """
        return len(self.forward)

    def __getitem__(self, n: FullIndex) -> NDArray[Any]:
        """Retrieves data from the forward dataset based on the given index.

        Parameters
        ----------
        n : Index
            Index specifying the data to retrieve.

        Returns
        -------
        Any
            Data from the forward dataset based on the index.
        """
        return self.forward[n]

    @property
    def name(self) -> Optional[str]:
        """Returns the name of the forward dataset."""
        if self._name is not None:
            return self._name
        return self.forward.name

    @property
    def dates(self) -> NDArray[np.datetime64]:
        """Returns the dates of the forward dataset."""
        return self.forward.dates

    @property
    def resolution(self) -> str:
        """Returns the resolution of the forward dataset."""
        return self.forward.resolution

    @property
    def field_shape(self) -> Shape:
        """Returns the field shape of the forward dataset."""
        return self.forward.field_shape

    @property
    def frequency(self) -> datetime.timedelta:
        """Returns the frequency of the forward dataset."""
        return self.forward.frequency

    @property
    def latitudes(self) -> NDArray[Any]:
        """Returns the latitudes of the forward dataset."""
        return self.forward.latitudes

    @property
    def longitudes(self) -> NDArray[Any]:
        """Returns the longitudes of the forward dataset."""
        return self.forward.longitudes

    @property
    def name_to_index(self) -> Dict[str, int]:
        """Returns a dictionary mapping variable names to their indices."""
        return self.forward.name_to_index

    @property
    def variables(self) -> List[str]:
        """Returns the variables of the forward dataset."""
        return self.forward.variables

    @property
    def variables_metadata(self) -> Dict[str, Any]:
        """Returns the metadata of the variables in the forward dataset."""
        return self.forward.variables_metadata

    @property
    def statistics(self) -> Dict[str, NDArray[Any]]:
        """Returns the statistics of the forward dataset."""
        return self.forward.statistics

    def statistics_tendencies(self, delta: Optional[datetime.timedelta] = None) -> Dict[str, NDArray[Any]]:
        """Returns the statistics tendencies of the forward dataset.

        Parameters
        ----------
        delta : Optional[Any]
            Time delta for calculating tendencies.

        Returns
        -------
        Any
            Statistics tendencies of the forward dataset.
        """
        if delta is None:
            delta = self.frequency
        return self.forward.statistics_tendencies(delta)

    @property
    def shape(self) -> Shape:
        """Returns the shape of the forward dataset."""
        return self.forward.shape

    @property
    def dtype(self) -> Any:
        """Returns the data type of the forward dataset."""
        return self.forward.dtype

    @property
    def missing(self) -> Set[int]:
        """Returns the missing data information of the forward dataset."""
        return self.forward.missing

    @property
    def grids(self) -> Any:
        """Returns the grids of the forward dataset."""
        return self.forward.grids

    def metadata_specific(self, **kwargs: Any) -> Dict[str, Any]:
        """Returns metadata specific to the forward dataset.

        Parameters
        ----------
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Dict[str, Any]
            Metadata specific to the forward dataset.
        """
        return super().metadata_specific(
            forward=self.forward.metadata_specific(),
            **self.forwards_subclass_metadata_specific(),
            **kwargs,
        )

    def collect_supporting_arrays(self, collected: List[Any], *path: Any) -> None:
        """Collects supporting arrays from the forward dataset.

        Parameters
        ----------
        collected : List[Any]
            List to which the supporting arrays are appended.
        *path : Any
            Variable length argument list specifying the paths for the arrays.
        """
        self.forward.collect_supporting_arrays(collected, *path)

    def collect_input_sources(self, collected: List[Any]) -> None:
        """Collects input sources from the forward dataset.

        Parameters
        ----------
        collected : List[Any]
            List to which the input sources are appended.
        """
        self.forward.collect_input_sources(collected)

    def source(self, index: int) -> Any:
        """Returns the source of the data at the specified index.

        Parameters
        ----------
        index : int
            Index specifying the data source.

        Returns
        -------
        Any
            Source of the data at the specified index.
        """
        return self.forward.source(index)

    @abstractmethod
    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Returns metadata specific to the subclass."""
        pass

    def get_dataset_names(self, names: Set[str]) -> None:
        """Collects the names of the datasets.

        Parameters
        ----------
        names : set
            Set to which the dataset names are added.
        """
        self.forward.get_dataset_names(names)

    @property
    def constant_fields(self) -> List[str]:
        """Returns the constant fields of the forward dataset."""
        return self.forward.constant_fields


class Combined(Forwards):
    """A class to combine multiple datasets into a single dataset."""

    def __init__(self, datasets: List[Dataset]) -> None:
        """Initializes a Combined object.

        Parameters
        ----------
        datasets : List[Dataset]
            List of datasets to be combined.
        """
        self.datasets = datasets
        assert len(self.datasets) > 1, len(self.datasets)

        for d in self.datasets[1:]:
            self.check_compatibility(self.datasets[0], d)

        # Forward most properties to the first dataset
        super().__init__(datasets[0])

    def mutate(self) -> Dataset:
        """Returns the mutated dataset.

        Returns
        -------
        Dataset
            Mutated dataset.
        """
        return self

    def check_same_resolution(self, d1: Dataset, d2: Dataset) -> None:
        """Checks if the resolutions of two datasets are the same.

        Parameters
        ----------
        d1 : Any
            First dataset.
        d2 : Any
            Second dataset.

        Raises
        ------
        ValueError
            If the resolutions are not the same.
        """
        if d1.resolution != d2.resolution:
            raise ValueError(f"Incompatible resolutions: {d1.resolution} and {d2.resolution} ({d1} {d2})")

    def check_same_frequency(self, d1: Dataset, d2: Dataset) -> None:
        """Checks if the frequencies of two datasets are the same.

        Parameters
        ----------
        d1 : Any
            First dataset.
        d2 : Any
            Second dataset.

        Raises
        ------
        ValueError
            If the frequencies are not the same.
        """
        if d1.frequency != d2.frequency:
            raise ValueError(f"Incompatible frequencies: {d1.frequency} and {d2.frequency} ({d1} {d2})")

    def check_same_grid(self, d1: Dataset, d2: Dataset) -> None:
        """Checks if the grids of two datasets are the same.

        Parameters
        ----------
        d1 : Any
            First dataset.
        d2 : Any
            Second dataset.

        Raises
        ------
        ValueError
            If the grids are not the same.
        """
        if (d1.latitudes != d2.latitudes).any() or (d1.longitudes != d2.longitudes).any():
            raise ValueError(f"Incompatible grid ({d1} {d2})")

    def check_same_shape(self, d1: Dataset, d2: Dataset) -> None:
        """Checks if the shapes of two datasets are the same.

        Parameters
        ----------
        d1 : Any
            First dataset.
        d2 : Any
            Second dataset.

        Raises
        ------
        ValueError
            If the shapes are not the same.
        """
        if d1.shape[1:] != d2.shape[1:]:
            raise ValueError(f"Incompatible shapes: {d1.shape} and {d2.shape} ({d1} {d2})")

        if d1.variables != d2.variables:
            raise ValueError(f"Incompatible variables: {d1.variables} and {d2.variables} ({d1} {d2})")

    def check_same_sub_shapes(self, d1: Any, d2: Any, drop_axis: int) -> None:
        """Checks if the sub-shapes of two datasets are the same along a given axis.

        Parameters
        ----------
        d1 : Any
            First dataset.
        d2 : Any
            Second dataset.
        drop_axis : int
            Axis along which to check the sub-shapes.

        Raises
        ------
        ValueError
            If the sub-shapes are not the same.
        """
        shape1 = d1.sub_shape(drop_axis)
        shape2 = d2.sub_shape(drop_axis)

        if shape1 != shape2:
            raise ValueError(f"Incompatible shapes: {d1.shape} and {d2.shape} ({d1} {d2})")

    def check_same_variables(self, d1: Dataset, d2: Dataset) -> None:
        """Checks if the variables of two datasets are the same.

        Parameters
        ----------
        d1 : Any
            First dataset.
        d2 : Any
            Second dataset.

        Raises
        ------
        ValueError
            If the variables are not the same.
        """
        if d1.variables != d2.variables:
            raise ValueError(f"Incompatible variables: {d1.variables} and {d2.variables} ({d1} {d2})")

    def check_same_lengths(self, d1: Dataset, d2: Dataset) -> None:
        """Checks if the lengths of two datasets are the same.

        Parameters
        ----------
        d1 : Any
            First dataset.
        d2 : Any
            Second dataset.

        Raises
        ------
        ValueError
            If the lengths are not the same.
        """
        if d1._len != d2._len:
            raise ValueError(f"Incompatible lengths: {d1._len} and {d2._len}")

    def check_same_dates(self, d1: Dataset, d2: Dataset) -> None:
        """Checks if the dates of two datasets are the same.

        Parameters
        ----------
        d1 : Any
            First dataset.
        d2 : Any
            Second dataset.

        Raises
        ------
        ValueError
            If the dates are not the same.
        """
        self.check_same_frequency(d1, d2)

        if d1.dates[0] != d2.dates[0]:
            raise ValueError(f"Incompatible start dates: {d1.dates[0]} and {d2.dates[0]} ({d1} {d2})")

        if d1.dates[-1] != d2.dates[-1]:
            raise ValueError(f"Incompatible end dates: {d1.dates[-1]} and {d2.dates[-1]} ({d1} {d2})")

    def check_compatibility(self, d1: Dataset, d2: Dataset) -> None:
        """Checks if two datasets are compatible.

        Parameters
        ----------
        d1 : Any
            First dataset.
        d2 : Any
            Second dataset.

        Raises
        ------
        ValueError
            If the datasets are not compatible.
        """
        # These are the default checks
        # Derived classes should turn individual checks off if they are not needed
        self.check_same_resolution(d1, d2)
        self.check_same_frequency(d1, d2)
        self.check_same_grid(d1, d2)
        self.check_same_lengths(d1, d2)
        self.check_same_variables(d1, d2)
        self.check_same_dates(d1, d2)

    def provenance(self) -> List[Any]:
        """Returns the provenance of the combined datasets.

        Returns
        -------
        List[Any]
            Provenance of the combined datasets.
        """
        return [d.provenance() for d in self.datasets]

    def __repr__(self) -> str:
        """Returns a string representation of the Combined object.

        Returns
        -------
        str
            String representation of the Combined object.
        """
        lst = ", ".join(repr(d) for d in self.datasets)
        return f"{self.__class__.__name__}({lst})"

    def metadata_specific(self, **kwargs: Any) -> Dict[str, Any]:
        """Returns metadata specific to the combined datasets.

        Parameters
        ----------
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Any
            Metadata specific to the combined datasets.
        """
        # We need to skip the forward superclass
        # TODO: revisit this
        return Dataset.metadata_specific(
            self,
            datasets=[d.metadata_specific() for d in self.datasets],
            **kwargs,
        )

    def collect_supporting_arrays(self, collected: List[Any], *path: Any) -> None:
        """Collects supporting arrays from the combined datasets.

        Parameters
        ----------
        collected : List[Any]
            List to which the supporting arrays are appended.
        *path : Any
            Variable length argument list specifying the paths for the arrays.
        """
        warnings.warn(f"The behaviour of {self.__class__.__name__}.collect_supporting_arrays() is not well defined")
        for i, d in enumerate(self.datasets):
            name = d.name if d.name is not None else i
            d.collect_supporting_arrays(collected, *path, name)

    @property
    def missing(self) -> Set[int]:
        """Returns the missing data information of the combined datasets.

        Raises
        ------
        NotImplementedError
            If the method is not implemented for Combined.
        """
        raise NotImplementedError("missing() not implemented for Combined")

    def get_dataset_names(self, names: Set[str]) -> None:
        """Collects the names of the combined datasets.

        Parameters
        ----------
        names : set
            Set to which the dataset names are added.
        """
        for d in self.datasets:
            d.get_dataset_names(names)


class GivenAxis(Combined):
    """A class to combine datasets along a given axis."""

    def __init__(self, datasets: List[Any], axis: int) -> None:
        """Initializes a GivenAxis object.

        Parameters
        ----------
        datasets : List[Any]
            List of datasets to be combined.
        axis : int
            Axis along which to combine the datasets.
        """
        self.axis = axis
        super().__init__(datasets)

        assert axis > 0 and axis < len(self.datasets[0].shape), (
            axis,
            self.datasets[0].shape,
        )

    def check_compatibility(self, d1: Dataset, d2: Dataset) -> None:
        """Checks if two datasets are compatible along the given axis.

        Parameters
        ----------
        d1 : Any
            First dataset.
        d2 : Any
            Second dataset.

        Raises
        ------
        ValueError
            If the datasets are not compatible along the given axis.
        """
        super().check_compatibility(d1, d2)
        self.check_same_sub_shapes(d1, d2, drop_axis=self.axis)

    @cached_property
    def shape(self) -> Shape:
        """Returns the shape of the combined dataset along the given axis."""
        shapes = [d.shape for d in self.datasets]
        before = shapes[0][: self.axis]
        after = shapes[0][self.axis + 1 :]
        result = before + (sum(s[self.axis] for s in shapes),) + after
        assert False not in result, result
        return result

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index: TupleIndex) -> NDArray[Any]:
        """Retrieves data from the combined dataset based on the given index.

        Parameters
        ----------
        index : Union[int, slice, Tuple[Union[int, slice], ...]]
            Index specifying the data to retrieve.

        Returns
        -------
        NDArray[Any]
            Data from the combined dataset based on the index.
        """
        index, changes = index_to_slices(index, self.shape)
        lengths = [d.shape[self.axis] for d in self.datasets]
        slices = length_to_slices(index[self.axis], lengths)
        result = [d[update_tuple(index, self.axis, i)[0]] for (d, i) in zip(self.datasets, slices) if i is not None]
        result = np.concatenate(result, axis=self.axis)
        return apply_index_to_slices_changes(result, changes)

    @debug_indexing
    def _get_slice(self, s: slice) -> NDArray[Any]:
        """Retrieves a slice of data from the combined dataset.

        Parameters
        ----------
        s : slice
            Slice specifying the data to retrieve.

        Returns
        -------
        NDArray[Any]
            Slice of data from the combined dataset.
        """
        return np.stack([self[i] for i in range(*s.indices(self._len))])

    @debug_indexing
    def __getitem__(self, n: FullIndex) -> NDArray[Any]:
        """Retrieves data from the combined dataset based on the given index.

        Parameters
        ----------
        n : Union[int, slice, Tuple[Union[int, slice], ...]]
            Index specifying the data to retrieve.

        Returns
        -------
        NDArray[Any]
            Data from the combined dataset based on the index.
        """
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        return np.concatenate([d[n] for d in self.datasets], axis=self.axis - 1)

    @cached_property
    def missing(self) -> Set[int]:
        """Returns the missing data information of the combined dataset along the given axis."""
        offset = 0
        result: Set[int] = set()
        for d in self.datasets:
            result.update(offset + m for m in d.missing)
            if self.axis == 0:  # Advance if axis is time
                offset += len(d)
        return result



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from anemoi.utils.dates import as_datetime

from .coordinates import Coordinate
from .variable import Variable

LOG = logging.getLogger(__name__)


class Time(ABC):
    """Base class for different time representations."""

    @classmethod
    def from_coordinates(cls, coordinates: List[Coordinate]) -> "Time":
        """Create a Time instance from a list of coordinates.

        Returns
        -------
        Union[ForecastFromValidTimeAndStep, Analysis, Constant, ForecastFromValidTimeAndBaseTime, ForecastFromBaseTimeAndDate]
            An instance of a subclass of Time.

        Args
        ----
        coordinates : List[Coordinate]
            List of coordinate objects.
        """
        time_coordinate = [c for c in coordinates if c.is_time]
        step_coordinate = [c for c in coordinates if c.is_step]
        date_coordinate = [c for c in coordinates if c.is_date]

        if len(date_coordinate) == 0 and len(time_coordinate) == 1 and len(step_coordinate) == 1:
            return ForecastFromValidTimeAndStep(time_coordinate[0], step_coordinate[0])

        if len(date_coordinate) == 0 and len(time_coordinate) == 1 and len(step_coordinate) == 0:
            return Analysis(time_coordinate[0])

        if len(date_coordinate) == 0 and len(time_coordinate) == 0 and len(step_coordinate) == 0:
            return Constant()

        if len(date_coordinate) == 1 and len(time_coordinate) == 1 and len(step_coordinate) == 0:
            return ForecastFromValidTimeAndBaseTime(date_coordinate[0], time_coordinate[0])

        if len(date_coordinate) == 1 and len(time_coordinate) == 0 and len(step_coordinate) == 1:
            return ForecastFromBaseTimeAndDate(date_coordinate[0], step_coordinate[0])

        if len(date_coordinate) == 1 and len(time_coordinate) == 1 and len(step_coordinate) == 1:
            return ForecastFromValidTimeAndStep(time_coordinate[0], step_coordinate[0], date_coordinate[0])

        LOG.error("")
        LOG.error(f"{len(date_coordinate)} date_coordinate")
        for c in date_coordinate:
            LOG.error("    %s %s %s %s", c, c.is_date, c.is_time, c.is_step)
            # LOG.error('    %s', c.variable)

        LOG.error("")
        LOG.error(f"{len(time_coordinate)} time_coordinate")
        for c in time_coordinate:
            LOG.error("    %s %s %s %s", c, c.is_date, c.is_time, c.is_step)
            # LOG.error('    %s', c.variable)

        LOG.error("")
        LOG.error(f"{len(step_coordinate)} step_coordinate")
        for c in step_coordinate:
            LOG.error("    %s %s %s %s", c, c.is_date, c.is_time, c.is_step)
            # LOG.error('    %s', c.variable)

        raise NotImplementedError(f"{len(date_coordinate)=} {len(time_coordinate)=} {len(step_coordinate)=}")

    @abstractmethod
    def select_valid_datetime(self, variable: Variable) -> Optional[str]:
        """Select the valid datetime for a given variable.

        Parameters
        ----------
        variable : Variable
            The variable to select the datetime for.

        Returns
        -------
        Optional[str]
            The name of the time coordinate.
        """
        pass

    @abstractmethod
    def fill_time_metadata(self, coords_values: Dict[str, Any], metadata: Dict[str, Any]) -> None:
        """Fill metadata with time information.

        Args
        ----
        coords_values : Dict[str, Any]
            Coordinate values.
        metadata : Dict[str, Any]
            Metadata dictionary.
        """

        pass


class Constant(Time):
    """Represents a constant time."""

    def fill_time_metadata(self, coords_values: Dict[str, Any], metadata: Dict[str, Any]) -> None:
        """Fill metadata with time information.

        Parameters
        ----------
        coords_values : Dict[str, Any]
            Coordinate values.
        metadata : Dict[str, Any]
            Metadata dictionary.
        """
        return None

    def select_valid_datetime(self, variable: Variable) -> None:
        """Select the valid datetime for a given variable.

        Parameters
        ----------
        variable : Variable
            The variable to select the datetime for.
        """
        return None


class Analysis(Time):
    """Represents an analysis time."""

    def __init__(self, time_coordinate: Coordinate) -> None:
        """Initialize Analysis with a time coordinate.

        Parameters
        ----------
        time_coordinate : Coordinate
            The time coordinate.
        """
        self.time_coordinate_name = time_coordinate.variable.name

    def fill_time_metadata(self, coords_values: Dict[str, Any], metadata: Dict[str, Any]) -> Any:
        """Fill metadata with time information.

        Parameters
        ----------
        coords_values : Dict[str, Any]
            Coordinate values.
        metadata : Dict[str, Any]
            Metadata dictionary.

        Returns
        -------
        Any
            The valid datetime.
        """
        valid_datetime = coords_values[self.time_coordinate_name]

        metadata["date"] = as_datetime(valid_datetime).strftime("%Y%m%d")
        metadata["time"] = as_datetime(valid_datetime).strftime("%H%M")
        metadata["step"] = 0

        return valid_datetime

    def select_valid_datetime(self, variable: Variable) -> str:
        """Select the valid datetime for a given variable.

        Parameters
        ----------
        variable : Variable
            The variable to select the datetime for.

        Returns
        -------
        str
            The name of the time coordinate.
        """
        return self.time_coordinate_name


class ForecastFromValidTimeAndStep(Time):
    """Represents a forecast time derived from valid time and step."""

    def __init__(
        self, time_coordinate: Coordinate, step_coordinate: Coordinate, date_coordinate: Optional[Coordinate] = None
    ) -> None:
        """Initialize ForecastFromValidTimeAndStep with time, step, and optional date coordinates.

        Args
        ----
        time_coordinate : Coordinate
            The time coordinate.
        step_coordinate : Coordinate
            The step coordinate.
        date_coordinate : Optional[Coordinate]
            The date coordinate.
        """
        self.time_coordinate_name = time_coordinate.variable.name
        self.step_coordinate_name = step_coordinate.variable.name
        self.date_coordinate_name = date_coordinate.variable.name if date_coordinate else None

    def fill_time_metadata(self, coords_values: Dict[str, Any], metadata: Dict[str, Any]) -> Any:
        """Fill metadata with time information.

        Returns
        -------
        Any
            The valid datetime.

        Args
        ----
        coords_values : Dict[str, Any]
            Coordinate values.
        metadata : Dict[str, Any]
            Metadata dictionary.
        """
        valid_datetime = coords_values[self.time_coordinate_name]
        step = coords_values[self.step_coordinate_name]

        assert isinstance(step, datetime.timedelta)
        base_datetime = valid_datetime - step

        hours = step.total_seconds() / 3600
        assert int(hours) == hours

        metadata["date"] = as_datetime(base_datetime).strftime("%Y%m%d")
        metadata["time"] = as_datetime(base_datetime).strftime("%H%M")
        metadata["step"] = int(hours)

        # When date is present, it should be compatible with time and step

        if self.date_coordinate_name is not None:
            # Not sure that this is the correct assumption
            assert coords_values[self.date_coordinate_name] == base_datetime, (
                coords_values[self.date_coordinate_name],
                base_datetime,
            )

        return valid_datetime

    def select_valid_datetime(self, variable: Variable) -> str:
        """Select the valid datetime for a given variable.

        Parameters
        ----------
        variable : Variable
            The variable to select the datetime for.

        Returns
        -------
        str
            The name of the time coordinate.
        """
        return self.time_coordinate_name


class ForecastFromValidTimeAndBaseTime(Time):
    """Represents a forecast time derived from valid time and base time."""

    def __init__(self, date_coordinate: Coordinate, time_coordinate: Coordinate) -> None:
        """Initialize ForecastFromValidTimeAndBaseTime with date and time coordinates.

        Args
        ----
        date_coordinate : Coordinate
            The date coordinate.
        time_coordinate : Coordinate
            The time coordinate.
        """
        self.date_coordinate_name = date_coordinate.name
        self.time_coordinate_name = time_coordinate.name

    def fill_time_metadata(self, coords_values: Dict[str, Any], metadata: Dict[str, Any]) -> Any:
        """Fill metadata with time information.

        Returns
        -------
        Any
            The valid datetime.

        Args
        ----
        coords_values : Dict[str, Any]
            Coordinate values.
        metadata : Dict[str, Any]
            Metadata dictionary.
        """
        valid_datetime = coords_values[self.time_coordinate_name]
        base_datetime = coords_values[self.date_coordinate_name]

        step = valid_datetime - base_datetime

        hours = step.total_seconds() / 3600
        assert int(hours) == hours

        metadata["date"] = as_datetime(base_datetime).strftime("%Y%m%d")
        metadata["time"] = as_datetime(base_datetime).strftime("%H%M")
        metadata["step"] = int(hours)

        return valid_datetime

    def select_valid_datetime(self, variable: Variable) -> str:
        """Select the valid datetime for a given variable.

        Parameters
        ----------
        variable : Variable
            The variable to select the datetime for.

        Returns
        -------
        str
            The name of the time coordinate.
        """
        return self.time_coordinate_name


class ForecastFromBaseTimeAndDate(Time):
    """Represents a forecast time derived from base time and date."""

    def __init__(self, date_coordinate: Coordinate, step_coordinate: Coordinate) -> None:
        """Initialize ForecastFromBaseTimeAndDate with date and step coordinates.

        Args
        ----
        date_coordinate : Coordinate
            The date coordinate.
        step_coordinate : Coordinate
            The step coordinate.
        """
        self.date_coordinate_name = date_coordinate.name
        self.step_coordinate_name = step_coordinate.name

    def fill_time_metadata(self, coords_values: Dict[str, Any], metadata: Dict[str, Any]) -> Any:
        """Fill metadata with time information.

        Returns
        -------
        Any
            The valid datetime.

        Args
        ----
        coords_values : Dict[str, Any]
            Coordinate values.
        metadata : Dict[str, Any]
            Metadata dictionary.
        """
        date = coords_values[self.date_coordinate_name]
        step = coords_values[self.step_coordinate_name]
        assert isinstance(step, datetime.timedelta)

        metadata["date"] = as_datetime(date).strftime("%Y%m%d")
        metadata["time"] = as_datetime(date).strftime("%H%M")

        hours = step.total_seconds() / 3600

        assert int(hours) == hours
        metadata["step"] = int(hours)

        return date + step

    def select_valid_datetime(self, variable: Variable) -> Optional[str]:
        """Select the valid datetime for a given variable.

        Parameters
        ----------
        variable : Variable
            The variable to select the datetime for.

        Returns
        -------
        Optional[str]
            The name of the time coordinate.
        """
        raise NotImplementedError("ForecastFromBaseTimeAndDate.select_valid_datetime")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Union

from earthkit.data.core.fieldlist import MultiFieldList

from anemoi.datasets.create.sources.mars import mars

from .legacy import legacy_source

LOGGER = logging.getLogger(__name__)


def _to_list(x: Union[list, tuple, Any]) -> List[Any]:
    """Converts the input to a list if it is not already a list or tuple.

    Parameters
    ----------
    x : Any
        The input to convert.

    Returns
    -------
    list
        A list containing the input elements.
    """
    if isinstance(x, (list, tuple)):
        return x
    return [x]


@legacy_source(__file__)
def hindcasts(context: Any, dates: List[Any], **request: Dict[str, Any]) -> MultiFieldList:
    """Generates hindcast requests based on the provided dates and request parameters.

    Parameters
    ----------
    context : Any
        The context containing the dates provider and trace method.
    dates : List[Any]
        A list of dates for which to generate hindcast requests.
    request : Dict[str, Any]
        Additional request parameters.

    Returns
    -------
    MultiFieldList
        A MultiFieldList containing the hindcast data.
    """
    from anemoi.datasets.dates import HindcastsDates

    provider = context.dates_provider
    assert isinstance(provider, HindcastsDates)

    context.trace("H️", f"hindcasts {len(dates)=}")

    request["param"] = _to_list(request["param"])
    request["step"] = _to_list(request.get("step", 0))
    request["step"] = [int(_) for _ in request["step"]]

    context.trace("H️", f"hindcast {request}")

    requests = []
    for d in dates:
        r = request.copy()
        hindcast = provider.mapping[d]
        r["hdate"] = hindcast.hdate.strftime("%Y-%m-%d")
        r["date"] = hindcast.refdate.strftime("%Y-%m-%d")
        r["time"] = hindcast.refdate.strftime("%H")
        r["step"] = hindcast.step
        requests.append(r)

    if len(requests) == 0:
        return MultiFieldList([])

    return mars(
        context,
        dates,
        *requests,
        date_key="hdate",
        request_already_using_valid_datetime=True,
    )


execute = hindcasts



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from collections import defaultdict
from typing import Any
from typing import Dict

import earthkit.data as ekd
from anemoi.transform.fields import new_field_from_numpy
from anemoi.transform.fields import new_fieldlist_from_list
from earthkit.meteo import thermo

from .legacy import legacy_filter


@legacy_filter(__file__)
def execute(context: Any, input: ekd.FieldList, t: str, td: str, rh: str = "d") -> ekd.FieldList:
    """Convert dewpoint on single levels to relative humidity.

    Parameters
    ----------
    context : Any
        The context in which the function is executed.
    input : List[Any]
        List of input fields.
    t : str
        Temperature parameter.
    td : str
        Dewpoint parameter.
    rh : str, optional
        Relative humidity parameter. Defaults to "d".

    Returns
    -------
    ekd.FieldList
        Array of fields with relative humidity.
    """
    result = []
    params: tuple[str, str] = (t, td)
    pairs: Dict[tuple, Dict[str, Any]] = defaultdict(dict)

    # Gather all necessary fields
    for f in input:
        key = f.metadata(namespace="mars")
        param = key.pop("param")
        if param in params:
            key = tuple(key.items())

            if param in pairs[key]:
                raise ValueError(f"Duplicate field {param} for {key}")

            pairs[key][param] = f
            if param == t:
                result.append(f)
        # all other parameters
        else:
            result.append(f)

    for keys, values in pairs.items():
        # some checks

        if len(values) != 2:
            raise ValueError("Missing fields")

        t_values = values[t].to_numpy(flatten=True)
        td_values = values[td].to_numpy(flatten=True)
        # actual conversion from td --> rh
        rh_values = thermo.relative_humidity_from_dewpoint(t=t_values, td=td_values)
        result.append(new_field_from_numpy(values[td], rh_values, param=rh))

    return new_fieldlist_from_list(result)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import calendar
import datetime
import logging
from pathlib import PurePath
from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import zarr
from anemoi.utils.config import load_config as load_settings
from numpy.typing import NDArray

if TYPE_CHECKING:
    from .dataset import Dataset

LOG = logging.getLogger(__name__)


def load_config() -> Dict[str, Any]:
    """Load the configuration settings.

    Returns
    -------
    Dict[str, Any]
        The configuration settings.
    """
    return load_settings(defaults={"datasets": {"named": {}, "path": []}})


def add_named_dataset(name: str, path: str, **kwargs: Any) -> None:
    """Add a named dataset to the configuration.

    Parameters
    ----------
    name : str
        The name of the dataset.
    path : str
        The path to the dataset.
    **kwargs : Any
        Additional arguments.
    """
    config = load_config()
    if name["datasets"]["named"]:
        raise ValueError(f"Dataset {name} already exists")

    config["datasets"]["named"][name] = path


def add_dataset_path(path: str) -> None:
    """Add a dataset path to the configuration.

    Parameters
    ----------
    path : str
        The path to add.
    """
    config = load_config()

    if path not in config["datasets"]["path"]:
        config["datasets"]["path"].append(path)


def round_datetime(d: np.datetime64, dates: NDArray[np.datetime64], up: bool) -> np.datetime64:
    """Round up (or down) a datetime to the nearest date in a list of dates.

    Parameters
    ----------
    d : np.datetime64
        The datetime to round.
    dates : NDArray[np.datetime64]
        The list of dates.
    up : bool
        Whether to round up.

    Returns
    -------
    np.datetime64
        The rounded datetime.
    """
    if dates is None or len(dates) == 0:
        return d

    for i, date in enumerate(dates):
        if date == d:
            return date
        if date > d:
            if up:
                return date
            if i > 0:
                return dates[i - 1]
            return date
    return dates[-1]


def _as_date(
    d: Union[int, str, np.datetime64, datetime.date], dates: NDArray[np.datetime64], last: bool
) -> np.datetime64:
    """Convert a date to a numpy datetime64 object, rounding to the nearest date in a list of dates.

    Parameters
    ----------
    d : Union[int, str, np.datetime64, datetime.date]
        The date to convert.
    dates : NDArray[np.datetime64]
        The list of dates.
    last : bool
        Whether to round to the last date.

    Returns
    -------
    np.datetime64
        The converted date.
    """
    # WARNING,  datetime.datetime is a subclass of datetime.date
    # so we need to check for datetime.datetime first

    if isinstance(d, (np.datetime64, datetime.datetime)):
        d = round_datetime(d, dates, up=not last)
        return np.datetime64(d)

    if isinstance(d, datetime.date):
        d = d.year * 10_000 + d.month * 100 + d.day

    try:
        d = int(d)
    except (ValueError, TypeError):
        pass

    if isinstance(d, int):
        if len(str(d)) == 4:
            year = d
            if last:
                return _as_date(np.datetime64(f"{year:04}-12-31T23:59:59"), dates, last)
            else:
                return _as_date(np.datetime64(f"{year:04}-01-01T00:00:00"), dates, last)

        if len(str(d)) == 6:
            year = d // 100
            month = d % 100
            if last:
                _, last_day = calendar.monthrange(year, month)
                return _as_date(np.datetime64(f"{year:04}-{month:02}-{last_day:02}T23:59:59"), dates, last)
            else:
                return _as_date(np.datetime64(f"{year:04}-{month:02}-01T00:00:00"), dates, last)

        if len(str(d)) == 8:
            year = d // 10000
            month = (d % 10000) // 100
            day = d % 100
            if last:
                return _as_date(np.datetime64(f"{year:04}-{month:02}-{day:02}T23:59:59"), dates, last)
            else:
                return _as_date(np.datetime64(f"{year:04}-{month:02}-{day:02}T00:00:00"), dates, last)

    if isinstance(d, str):

        def isfloat(s: str) -> bool:
            try:
                float(s)
                return True
            except ValueError:
                return False

        if d.endswith("%") and isfloat(d[:-1]):
            x = float(d[:-1])
            if not 0 <= x <= 100:
                raise ValueError(f"Invalid date: {d}")
            i_float = x * len(dates) / 100

            epsilon = 2 ** (-30)
            if len(dates) > 1 / epsilon:
                LOG.warning("Too many dates to use percentage, one date may be lost in rounding")

            if last:
                index = int(i_float + epsilon) - 1
            else:
                index = int(i_float - epsilon)
            index = max(0, min(len(dates) - 1, index))
            return dates[index]

        if "-" in d and ":" in d:
            date, time = d.replace(" ", "T").split("T")
            year, month, day = [int(_) for _ in date.split("-")]
            hour, minute, second = [int(_) for _ in time.split(":")]
            return _as_date(
                np.datetime64(f"{year:04}-{month:02}-{day:02}T{hour:02}:{minute:02}:{second:02}"),
                dates,
                last,
            )

        if "-" in d:
            assert ":" not in d
            bits = d.split("-")
            if len(bits) == 1:
                return _as_date(int(bits[0]), dates, last)

            if len(bits) == 2:
                return _as_date(int(bits[0]) * 100 + int(bits[1]), dates, last)

            if len(bits) == 3:
                return _as_date(int(bits[0]) * 10000 + int(bits[1]) * 100 + int(bits[2]), dates, last)

        if ":" in d:
            assert len(d) == 5
            hour, minute = d.split(":")
            assert minute == "00"
            assert not last
            first = dates[0].astype(object)
            year = first.year
            month = first.month
            day = first.day

            return _as_date(np.datetime64(f"{year:04}-{month:02}-{day:02}T{hour}:00:00"), dates, last)

    raise NotImplementedError(f"Unsupported date: {d} ({type(d)})")


def as_first_date(d: Union[int, str, np.datetime64, datetime.date], dates: NDArray[np.datetime64]) -> np.datetime64:
    """Convert a date to the first date in a list of dates.

    Parameters
    ----------
    d : Union[int, str, np.datetime64, datetime.date]
        The date to convert.
    dates : NDArray[np.datetime64]
        The list of dates.

    Returns
    -------
    np.datetime64
        The first date.
    """
    return _as_date(d, dates, last=False)


def as_last_date(d: Union[int, str, np.datetime64, datetime.date], dates: NDArray[np.datetime64]) -> np.datetime64:
    """Convert a date to the last date in a list of dates.

    Parameters
    ----------
    d : Union[int, str, np.datetime64, datetime.date]
        The date to convert.
    dates : NDArray[np.datetime64]
        The list of dates.

    Returns
    -------
    np.datetime64
        The last date.
    """
    return _as_date(d, dates, last=True)


def _concat_or_join(datasets: List["Dataset"], kwargs: Dict[str, Any]) -> Tuple["Dataset", Dict[str, Any]]:
    """Concatenate or join datasets based on their date ranges.

    Parameters
    ----------
    datasets : List[Dataset]
        The list of datasets.
    kwargs : Dict[str, Any]
        Additional arguments.

    Returns
    -------
    Tuple[Dataset, Dict[str, Any]]
        The concatenated or joined dataset and remaining arguments.
    """
    if "adjust" in kwargs:
        raise ValueError("Cannot use 'adjust' without specifying 'concat' or 'join'")
    datasets, kwargs = _auto_adjust(datasets, kwargs)

    # Study the dates
    ranges = [(d.dates[0].astype(object), d.dates[-1].astype(object)) for d in datasets]

    if len(set(ranges)) == 1:
        from .join import Join

        return Join(datasets)._overlay(), kwargs

    from .concat import Concat

    Concat.check_dataset_compatibility(datasets)

    return Concat(datasets), kwargs


def _open(a: Union[str, PurePath, Dict[str, Any], List[Any], Tuple[Any, ...]]) -> "Dataset":
    """Open a dataset from various input types.

    Parameters
    ----------
    a : Union[str, PurePath, Dict[str, Any], List[Any], Tuple[Any, ...]]
        The input to open.

    Returns
    -------
    Dataset
        The opened dataset.
    """
    from .dataset import Dataset
    from .stores import Zarr
    from .stores import zarr_lookup

    if isinstance(a, Dataset):
        return a.mutate()

    if isinstance(a, zarr.hierarchy.Group):
        return Zarr(a).mutate()

    if isinstance(a, str):
        return Zarr(zarr_lookup(a)).mutate()

    if isinstance(a, PurePath):
        return _open(str(a)).mutate()

    if isinstance(a, dict):
        return _open_dataset(**a).mutate()

    if isinstance(a, (list, tuple)):
        return _open_dataset(*a).mutate()

    raise NotImplementedError(f"Unsupported argument: {type(a)}")


def _auto_adjust(
    datasets: List["Dataset"],
    kwargs: Dict[str, Any],
    exclude: Optional[List[str]] = None,
) -> Tuple[List["Dataset"], Dict[str, Any]]:
    """Automatically adjust datasets based on specified criteria.

    Parameters
    ----------
    datasets : List[Dataset]
        The list of datasets.
    kwargs : Dict[str, Any]
        Additional arguments.
    exclude : Optional[List[str]]
        List of keys to exclude from adjustment.

    Returns
    -------
    Tuple[List[Dataset], Dict[str, Any]]
        The adjusted datasets and remaining arguments.
    """
    if "adjust" not in kwargs:
        return datasets, kwargs

    adjust_list = kwargs.pop("adjust")
    if not isinstance(adjust_list, (tuple, list)):
        adjust_list = [adjust_list]

    ALIASES = {
        "all": ["select", "frequency", "start", "end"],
        "dates": ["start", "end", "frequency"],
        "variables": ["select"],
    }

    adjust_set = set()

    for a in adjust_list:
        adjust_set.update(ALIASES.get(a, [a]))

    if exclude is not None:
        adjust_set -= set(exclude)

    extra = set(adjust_set) - set(ALIASES["all"])
    if extra:
        raise ValueError(f"Invalid adjust keys: {extra}")

    subset_kwargs = [{} for _ in datasets]

    if "select" in adjust_set:
        assert "select" not in kwargs, "Cannot use 'select' in adjust and kwargs"

        variables = None

        for d in datasets:
            if variables is None:
                variables = set(d.variables)
            else:
                variables &= set(d.variables)

        if len(variables) == 0:
            raise ValueError("No common variables")

        for i, d in enumerate(datasets):
            if set(d.variables) != variables:
                subset_kwargs[i]["select"] = sorted(variables)

    if "start" or "end" in adjust_set:
        common = datasets[0].dates
        for d in datasets[0:]:
            common = np.intersect1d(common, d.dates)

    if "start" in adjust_set:
        assert "start" not in kwargs, "Cannot use 'start' in adjust and kwargs"
        start = min(common).astype(object)
        for i, d in enumerate(datasets):
            if start != d.dates[0]:
                subset_kwargs[i]["start"] = start

    if "end" in adjust_set:
        assert "end" not in kwargs, "Cannot use 'end' in adjust and kwargs"
        end = max(common).astype(object)
        for i, d in enumerate(datasets):
            if end != d.dates[-1]:
                subset_kwargs[i]["end"] = end

    if "frequency" in adjust_set:
        assert "frequency" not in kwargs, "Cannot use 'frequency' in adjust and kwargs"
        frequency = max(d.frequency for d in datasets)
        for i, d in enumerate(datasets):
            if d.frequency != frequency:
                subset_kwargs[i]["frequency"] = frequency

    datasets = [d._subset(**subset_kwargs[i]) for i, d in enumerate(datasets)]

    return datasets, kwargs


def _open_dataset(*args: Any, **kwargs: Any) -> "Dataset":
    """Open a dataset.

    Parameters
    ----------
    *args : Any
        Positional arguments.
    **kwargs : Any
        Keyword arguments.

    Returns
    -------
    Dataset
        The opened dataset.
    """
    sets = []
    for a in args:
        sets.append(_open(a))

    if "xy" in kwargs:
        # Experimental feature, may be removed
        from .xy import xy_factory

        assert not sets, sets
        return xy_factory(args, kwargs).mutate()

    if "x" in kwargs and "y" in kwargs:
        # Experimental feature, may be removed
        from .xy import xy_factory

        assert not sets, sets
        return xy_factory(args, kwargs).mutate()

    if "zip" in kwargs:
        # Experimental feature, may be removed
        from .xy import zip_factory

        assert not sets, sets
        return zip_factory(args, kwargs).mutate()

    if "chain" in kwargs:
        # Experimental feature, may be removed
        from .unchecked import chain_factory

        assert not sets, sets
        return chain_factory(args, kwargs).mutate()

    if "join" in kwargs:
        from .join import join_factory

        assert not sets, sets
        return join_factory(args, kwargs).mutate()

    if "concat" in kwargs:
        from .concat import concat_factory

        assert not sets, sets
        return concat_factory(args, kwargs).mutate()

    if "merge" in kwargs:
        from .merge import merge_factory

        assert not sets, sets
        return merge_factory(args, kwargs).mutate()

    if "ensemble" in kwargs:
        from .ensemble import ensemble_factory

        assert not sets, sets
        return ensemble_factory(args, kwargs).mutate()

    if "grids" in kwargs:
        from .grids import grids_factory

        assert not sets, sets
        return grids_factory(args, kwargs).mutate()

    if "cutout" in kwargs:
        from .grids import cutout_factory

        assert not sets, sets
        return cutout_factory(args, kwargs).mutate()

    if "complement" in kwargs:
        from .complement import complement_factory

        assert not sets, sets
        return complement_factory(args, kwargs).mutate()

    for name in ("datasets", "dataset"):
        if name in kwargs:
            datasets = kwargs.pop(name)
            if not isinstance(datasets, (list, tuple)):
                datasets = [datasets]
            for a in datasets:
                sets.append(_open(a))

    assert len(sets) > 0, (args, kwargs)

    if len(sets) > 1:
        dataset, kwargs = _concat_or_join(sets, kwargs)
        return dataset._subset(**kwargs)

    return sets[0]._subset(**kwargs)


def append_to_zarr(new_data: np.ndarray, new_dates: np.ndarray, zarr_path: str) -> None:
    """Append data from a subset (for one date) to the Zarr store.

    Parameters
    ----------
    new_data : np.ndarray
        The new data to append.
    new_dates : np.ndarray
        The new dates to append.
    zarr_path : str
        The path to the Zarr store.

    Notes
    -----
    - "dates" dataset is created with chunks equal to len(big_dataset.dates).
    - "data" dataset is created with chunk size 1 along the first (time) dimension.
    """
    print("Appending data for", new_dates, flush=True)
    # Re-open the zarr store to avoid root object accumulating memory.
    root = zarr.open(zarr_path, mode="a")
    # Convert new dates to strings (using str) regardless of input dtype.
    new_dates = np.array(new_dates, dtype="datetime64[ns]")
    dates_ds = root["dates"]
    old_len = dates_ds.shape[0]
    dates_ds.resize((old_len + len(new_dates),))
    dates_ds[old_len:] = new_dates
    # Append to "data" dataset.
    data_ds = root["data"]
    old_shape = data_ds.shape  # (time, n_vars, ensembles, gridpoints)
    new_shape = (old_shape[0] + len(new_dates),) + old_shape[1:]
    data_ds.resize(new_shape)
    data_ds[old_shape[0] :] = new_data


def process_date(date: Any, big_dataset: Any) -> Tuple[np.ndarray, np.ndarray]:
    """Open the subset corresponding to the given date and return (date, subset).

    Parameters
    ----------
    date : Any
        The date to process.
    big_dataset : Any
        The dataset to process.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The subset and the date.
    """
    print("Processing:", date, flush=True)
    subset = _open_dataset(big_dataset, start=date, end=date).mutate()
    s = subset[:]
    date = subset.dates
    return s, date


def initialize_zarr_store(root: Any, big_dataset: Any, recipe: Dict[str, Any]) -> None:
    """Initialize the Zarr store with the given dataset and recipe.

    Parameters
    ----------
    root : Any
        The root of the Zarr store.
    big_dataset : Any
        The dataset to initialize the store with.
    recipe : Dict[str, Any]
        The recipe for initializing the store.
    """
    ensembles = big_dataset.shape[1]
    # Create or append to "dates" dataset.
    if "dates" not in root:
        full_length = len(big_dataset.dates)
        root.create_dataset("dates", data=np.array([], dtype="datetime64[s]"), chunks=(full_length,))

    if "data" not in root:
        dims = (1, len(big_dataset.variables), ensembles, big_dataset.grids[0])
        root.create_dataset(
            "data",
            shape=dims,
            dtype=np.float64,
            chunks=dims,
        )

    for k, v in big_dataset.statistics.items():
        if k not in root:
            root.create_dataset(
                k,
                data=v,
                compressor=None,
            )

    # Create spatial coordinate datasets if missing.
    if "latitudes" not in root or "longitudes" not in root:
        root.create_dataset("latitudes", data=big_dataset.latitudes, compressor=None)
        root.create_dataset("longitudes", data=big_dataset.longitudes, compressor=None)

    # Set store-wide attributes if not already set.
    if "frequency" not in root.attrs:
        root.attrs["frequency"] = "10m"
        root.attrs["resolution"] = "1km"
        root.attrs["name_to_index"] = {k: i for i, k in enumerate(big_dataset.variables)}
        root.attrs["ensemble_dimension"] = 1
        root.attrs["field_shape"] = big_dataset.field_shape
        root.attrs["flatten_grid"] = True
        root.attrs["recipe"] = recipe


def _save_dataset(recipe: Dict[str, Any], zarr_path: str, n_workers: int = 1) -> None:
    """Incrementally create (or update) a Zarr store from an Anemoi dataset.

    Parameters
    ----------
    recipe : Dict[str, Any]
        The recipe for creating the dataset.
    zarr_path : str
        The path to the Zarr store.
    n_workers : int, optional
        The number of worker processes to use, by default 1.

    Notes
    -----
    Worker processes extract data for each date in parallel, but all writes
    to the store happen sequentially in the main process (i.e. single-writer).

    The "dates" dataset is created with chunking equal to the full length of
    big_dataset.dates, while "data" is chunked with 1 in the time dimension.
    """
    from concurrent.futures import ProcessPoolExecutor

    full_ds = _open_dataset(recipe).mutate()
    print("Opened full dataset.", flush=True)

    # Use ProcessPoolExecutor for parallel data extraction.
    # Workers return (date, subset) tuples.
    root = zarr.open(zarr_path, mode="a")
    initialize_zarr_store(root, full_ds, recipe)
    print("Zarr store initialized.", flush=True)

    existing_dates = np.array(sorted(root["dates"]), dtype="datetime64[s]")
    all_dates = full_ds.dates
    # To resume creation of the Zarr store in case the job is interrupted.
    dates_to_process = np.array(sorted(set(all_dates).difference(existing_dates)), dtype="datetime64[s]")

    use_pool = False

    if use_pool:
        with ProcessPoolExecutor(n_workers) as pool:
            futures = [pool.submit(process_date, date, full_ds) for date in dates_to_process]
            for future in futures:
                subset, date = future.result()
                # All appends happen sequentially here to
                #  avoid dates being added in a random order.
                append_to_zarr(subset, date, zarr_path)
    else:
        for date in dates_to_process:
            subset, date = process_date(date, full_ds)
            append_to_zarr(subset, date, zarr_path)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from copy import deepcopy
from typing import Any
from typing import Union

from anemoi.datasets.dates.groups import GroupOfDates

from .trace import trace_select

LOG = logging.getLogger(__name__)


class Context:
    """Context for building input data."""

    pass


class InputBuilder:
    """Builder class for creating input data from configuration and data sources."""

    def __init__(self, config: dict, data_sources: Union[dict, list], **kwargs: Any) -> None:
        """Initialize the InputBuilder.

        Parameters
        ----------
        config : dict
            Configuration dictionary.
        data_sources : Union[dict, list]
            Data sources.
        **kwargs : Any
            Additional keyword arguments.
        """
        self.kwargs = kwargs

        config = deepcopy(config)
        if data_sources:
            config = dict(
                data_sources=dict(
                    sources=data_sources,
                    input=config,
                )
            )
        self.config = config
        self.action_path = ["input"]

    @trace_select
    def select(self, group_of_dates: GroupOfDates) -> Any:
        """Select data based on the group of dates.

        Parameters
        ----------
        group_of_dates : GroupOfDates
            Group of dates to select data for.

        Returns
        -------
        Any
            Selected data.
        """
        from .action import ActionContext
        from .action import action_factory

        """This changes the context."""
        context = ActionContext(**self.kwargs)
        action = action_factory(self.config, context, self.action_path)
        return action.select(group_of_dates)

    def __repr__(self) -> str:
        """Return a string representation of the InputBuilder.

        Returns
        -------
        str
            String representation.
        """
        from .action import ActionContext
        from .action import action_factory

        context = ActionContext(**self.kwargs)
        a = action_factory(self.config, context, self.action_path)
        return repr(a)

    def _trace_select(self, group_of_dates: GroupOfDates) -> str:
        """Trace the select operation.

        Parameters
        ----------
        group_of_dates : GroupOfDates
            Group of dates to select data for.

        Returns
        -------
        str
            Trace string.
        """
        return f"InputBuilder({group_of_dates})"


def build_input(config: dict, data_sources: Union[dict, list], **kwargs: Any) -> InputBuilder:
    """Build an InputBuilder instance.

    Parameters
    ----------
    config : dict
        Configuration dictionary.
    data_sources : Union[dict, list]
        Data sources.
    **kwargs : Any
        Additional keyword arguments.

    Returns
    -------
    InputBuilder
        An instance of InputBuilder.
    """
    return InputBuilder(config, data_sources, **kwargs)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from . import source_registry
from .xarray import XarraySourceBase


@source_registry.register("xarray_kerchunk")
class XarrayKerchunkSource(XarraySourceBase):
    """An Xarray data source that uses the `kerchunk` engine."""

    emoji = "🧱"

    def __init__(self, context, json, *args, **kwargs: dict):
        super().__init__(context, *args, **kwargs)

        self.path_or_url = "reference://"

        self.options = {
            "engine": "zarr",
            "backend_kwargs": {
                "consolidated": False,
                "storage_options": {
                    "fo": json,
                    "remote_protocol": "s3",
                    "remote_options": {"anon": True},
                },
            },
        }



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
import re
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Union

from anemoi.utils.humanize import did_you_mean
from earthkit.data import from_source
from earthkit.data.utils.availability import Availability

from anemoi.datasets.create.utils import to_datetime_list

from .legacy import legacy_source

DEBUG = False


def to_list(x: Union[list, tuple, Any]) -> list:
    """Converts the input to a list if it is not already a list or tuple.

    Parameters
    ----------
    x : Any
        The input value to be converted.

    Returns
    -------
    list
        A list containing the input value(s).
    """
    if isinstance(x, (list, tuple)):
        return x
    return [x]


def _date_to_datetime(
    d: Union[datetime.datetime, list, tuple, str],
) -> Union[datetime.datetime, List[datetime.datetime]]:
    """Converts the input date(s) to datetime objects.

    Parameters
    ----------
    d : Union[datetime.datetime, list, tuple, str]
        The input date(s) to be converted.

    Returns
    -------
    Union[datetime.datetime, List[datetime.datetime]]
        A datetime object or a list of datetime objects.
    """
    if isinstance(d, datetime.datetime):
        return d
    if isinstance(d, (list, tuple)):
        return [_date_to_datetime(x) for x in d]
    return datetime.datetime.fromisoformat(d)


def expand_to_by(x: Union[str, int, list]) -> Union[str, int, list]:
    """Expands a range expression to a list of values.

    Parameters
    ----------
    x : Union[str, int, list]
        The input range expression.

    Returns
    -------
    Union[str, int, list]
        A list of expanded values.
    """
    if isinstance(x, (str, int)):
        return expand_to_by(str(x).split("/"))

    if len(x) == 3 and x[1] == "to":
        start = int(x[0])
        end = int(x[2])
        return list(range(start, end + 1))

    if len(x) == 5 and x[1] == "to" and x[3] == "by":
        start = int(x[0])
        end = int(x[2])
        by = int(x[4])
        return list(range(start, end + 1, by))

    return x


def normalise_time_delta(t: Union[datetime.timedelta, str]) -> datetime.timedelta:
    """Normalizes a time delta string to a datetime.timedelta object.

    Parameters
    ----------
    t : Union[datetime.timedelta, str]
        The input time delta string.

    Returns
    -------
    datetime.timedelta
        A normalized datetime.timedelta object.
    """
    if isinstance(t, datetime.timedelta):
        assert t == datetime.timedelta(hours=t.hours), t

    assert t.endswith("h"), t

    t = int(t[:-1])
    t = datetime.timedelta(hours=t)
    return t


def _normalise_time(t: Union[int, str]) -> str:
    """Normalizes a time value to a string in HHMM format.

    Parameters
    ----------
    t : Union[int, str]
        The input time value.

    Returns
    -------
    str
        A string representing the normalized time.
    """
    t = int(t)
    if t < 100:
        t * 100
    return "{:04d}".format(t)


def _expand_mars_request(
    request: Dict[str, Any],
    date: datetime.datetime,
    request_already_using_valid_datetime: bool = False,
    date_key: str = "date",
) -> List[Dict[str, Any]]:
    """Expands a MARS request with the given date and other parameters.

    Parameters
    ----------
    request : Dict[str, Any]
        The input MARS request.
    date : datetime.datetime
        The date to be used in the request.
    request_already_using_valid_datetime : bool, optional
        Flag indicating if the request already uses valid datetime.
    date_key : str, optional
        The key for the date in the request.

    Returns
    -------
    List[Dict[str, Any]]
        A list of expanded MARS requests.
    """
    requests = []

    user_step = to_list(expand_to_by(request.get("step", [0])))
    user_time = None
    user_date = None

    if not request_already_using_valid_datetime:
        user_time = request.get("time")
        if user_time is not None:
            user_time = to_list(user_time)
            user_time = [_normalise_time(t) for t in user_time]

        user_date = request.get(date_key)
        if user_date is not None:
            assert isinstance(user_date, str), user_date
            user_date = re.compile("^{}$".format(user_date.replace("-", "").replace("?", ".")))

    for step in user_step:
        r = request.copy()

        if not request_already_using_valid_datetime:

            if isinstance(step, str) and "-" in step:
                assert step.count("-") == 1, step

            # this takes care of the cases where the step is a period such as 0-24 or 12-24
            hours = int(str(step).split("-")[-1])

            base = date - datetime.timedelta(hours=hours)
            r.update(
                {
                    date_key: base.strftime("%Y%m%d"),
                    "time": base.strftime("%H%M"),
                    "step": step,
                }
            )

        for pproc in ("grid", "rotation", "frame", "area", "bitmap", "resol"):
            if pproc in r:
                if isinstance(r[pproc], (list, tuple)):
                    r[pproc] = "/".join(str(x) for x in r[pproc])

        if user_date is not None:
            if not user_date.match(r[date_key]):
                continue

        if user_time is not None:
            # It time is provided by the user, we only keep the requests that match the time
            if r["time"] not in user_time:
                continue

        requests.append(r)

    # assert requests, requests

    return requests


def factorise_requests(
    dates: List[datetime.datetime],
    *requests: Dict[str, Any],
    request_already_using_valid_datetime: bool = False,
    date_key: str = "date",
) -> Generator[Dict[str, Any], None, None]:
    """Factorizes the requests based on the given dates.

    Parameters
    ----------
    dates : List[datetime.datetime]
        The list of dates to be used in the requests.
    requests : Dict[str, Any]
        The input requests to be factorized.
    request_already_using_valid_datetime : bool, optional
        Flag indicating if the requests already use valid datetime.
    date_key : str, optional
        The key for the date in the requests.

    Returns
    -------
    Generator[Dict[str, Any], None, None]
        Factorized requests.
    """
    updates = []
    for req in requests:
        # req = normalise_request(req)

        for d in dates:
            updates += _expand_mars_request(
                req,
                date=d,
                request_already_using_valid_datetime=request_already_using_valid_datetime,
                date_key=date_key,
            )

    if not updates:
        return

    compressed = Availability(updates)
    for r in compressed.iterate():
        for k, v in r.items():
            if isinstance(v, (list, tuple)) and len(v) == 1:
                r[k] = v[0]
        yield r


def use_grib_paramid(r: Dict[str, Any]) -> Dict[str, Any]:
    """Converts the parameter short names to GRIB parameter IDs.

    Parameters
    ----------
    r : Dict[str, Any]
        The input request containing parameter short names.

    Returns
    -------
    Dict[str, Any]
        The request with parameter IDs.
    """
    from anemoi.utils.grib import shortname_to_paramid

    params = r["param"]
    if isinstance(params, str):
        params = params.split("/")
    assert isinstance(params, (list, tuple)), params

    params = [shortname_to_paramid(p) for p in params]
    r["param"] = "/".join(str(p) for p in params)

    return r


MARS_KEYS = [
    "accuracy",
    "activity",
    "anoffset",
    "area",
    "bitmap",
    "channel",
    "class",
    "database",
    "dataset",
    "date",
    "diagnostic",
    "direction",
    "domain",
    "expect",
    "experiment",
    "expver",
    "fcmonth",
    "fcperiod",
    "fieldset",
    "filter",
    "format",
    "frame",
    "frequency",
    "gaussian",
    "generation",
    "grid",
    "hdate",
    "ident",
    "instrument",
    "interpolation",
    "intgrid",
    "iteration",
    "level",
    "levelist",
    "levtype",
    "method",
    "model",
    "month",
    "number",
    "obsgroup",
    "obstype",
    "offsetdate",
    "offsettime",
    "optimise",
    "origin",
    "packing",
    "padding",
    "param",
    "quantile",
    "realization",
    "reference",
    "reportype",
    "repres",
    "resol",
    "resolution",
    "rotation",
    "step",
    "stream",
    "system",
    "target",
    "time",
    "truncation",
    "type",
    "year",
]


@legacy_source(__file__)
def mars(
    context: Any,
    dates: List[datetime.datetime],
    *requests: Dict[str, Any],
    request_already_using_valid_datetime: bool = False,
    date_key: str = "date",
    use_cdsapi_dataset: Optional[str] = None,
    **kwargs: Any,
) -> Any:
    """Executes MARS requests based on the given context, dates, and other parameters.

    Parameters
    ----------
    context : Any
        The context for the requests.
    dates : List[datetime.datetime]
        The list of dates to be used in the requests.
    requests : Dict[str, Any]
        The input requests to be executed.
    request_already_using_valid_datetime : bool, optional
        Flag indicating if the requests already use valid datetime.
    date_key : str, optional
        The key for the date in the requests.
    use_cdsapi_dataset : Optional[str], optional
        The dataset to be used with CDS API.
    kwargs : Any
        Additional keyword arguments for the requests.

    Returns
    -------
    Any
        The resulting dataset.
    """

    if not requests:
        requests = [kwargs]

    for r in requests:
        param = r.get("param", [])
        if not isinstance(param, (list, tuple)):
            param = [param]
        # check for "Norway bug" where yaml transforms 'no' into False, etc.
        for p in param:
            if p is False:
                raise ValueError(
                    "'param' cannot be 'False'. If you wrote 'param: no' or 'param: off' in yaml, you may want to use quotes?"
                )
            if p is None:
                raise ValueError(
                    "'param' cannot be 'None'. If you wrote 'param: no' in yaml, you may want to use quotes?"
                )
            if p is True:
                raise ValueError(
                    "'param' cannot be 'True'. If you wrote 'param: on' in yaml, you may want to use quotes?"
                )

    if len(dates) == 0:  # When using `repeated_dates`
        assert len(requests) == 1, requests
        assert "date" in requests[0], requests[0]
        if isinstance(requests[0]["date"], datetime.date):
            requests[0]["date"] = requests[0]["date"].strftime("%Y%m%d")
    else:
        requests = factorise_requests(
            dates,
            *requests,
            request_already_using_valid_datetime=request_already_using_valid_datetime,
            date_key=date_key,
        )

    requests = list(requests)

    ds = from_source("empty")
    context.trace("✅", f"{[str(d) for d in dates]}")
    context.trace("✅", f"Will run {len(requests)} requests")
    for r in requests:
        r = {k: v for k, v in r.items() if v != ("-",)}
        context.trace("✅", f"mars {r}")

    for r in requests:
        r = {k: v for k, v in r.items() if v != ("-",)}

        if context.use_grib_paramid and "param" in r:
            r = use_grib_paramid(r)

        for k, v in r.items():
            if k not in MARS_KEYS:
                raise ValueError(
                    f"⚠️ Unknown key {k}={v} in MARS request. Did you mean '{did_you_mean(k, MARS_KEYS)}' ?"
                )
        try:
            if use_cdsapi_dataset:
                ds = ds + from_source("cds", use_cdsapi_dataset, r)
            else:
                ds = ds + from_source("mars", **r)
        except Exception as e:
            if "File is empty:" not in str(e):
                raise
    return ds


execute = mars


if __name__ == "__main__":
    import yaml

    config = yaml.safe_load(
        """
    - class: ea
      expver: '0001'
      grid: 20.0/20.0
      levtype: sfc
      param: [2t]
      # param: [10u, 10v, 2d, 2t, lsm, msl, sdor, skt, slor, sp, tcw, z]
      number: [0, 1]

    # - class: ea
    #   expver: '0001'
    #   grid: 20.0/20.0
    #   levtype: pl
    #   param: [q]
    #   levelist: [1000, 850]

    """
    )
    dates = yaml.safe_load("[2022-12-30 18:00, 2022-12-31 00:00, 2022-12-31 06:00, 2022-12-31 12:00]")
    dates = to_datetime_list(dates)

    DEBUG = True
    for f in mars(None, dates, *config):
        print(f, f.to_numpy().mean())



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import warnings
from functools import reduce
from math import gcd
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# from anemoi.utils.dates import as_datetime
from anemoi.utils.dates import DateTimes
from anemoi.utils.dates import as_datetime
from anemoi.utils.dates import frequency_to_string
from anemoi.utils.dates import frequency_to_timedelta
from anemoi.utils.hindcasts import HindcastDatesTimes
from anemoi.utils.humanize import print_dates


def extend(x: Union[str, List[Any], Tuple[Any, ...]]) -> Iterator[datetime.datetime]:
    """Extend a date range or list of dates into individual datetime objects.

    Args:
        x (Union[str, List[Any], Tuple[Any, ...]]): A date range string or list/tuple of dates.

    Returns
    -------
    Iterator[datetime.datetime]
        An iterator of datetime objects.
    """

    if isinstance(x, (list, tuple)):
        for y in x:
            yield from extend(y)
        return

    if isinstance(x, str):
        if "/" in x:
            start, end, step = x.split("/")
            start = as_datetime(start)
            end = as_datetime(end)
            step = frequency_to_timedelta(step)
            while start <= end:
                yield start
                start += step
            return

    yield as_datetime(x)


class DatesProvider:
    """Base class for date generation.

    >>> DatesProvider.from_config(**{"start": "2023-01-01 00:00", "end": "2023-01-02 00:00", "frequency": "1d"}).values
    [datetime.datetime(2023, 1, 1, 0, 0), datetime.datetime(2023, 1, 2, 0, 0)]

    >>> DatesProvider.from_config(**{"start": "2023-01-01 00:00", "end": "2023-01-03 00:00", "frequency": "18h"}).values
    [datetime.datetime(2023, 1, 1, 0, 0), datetime.datetime(2023, 1, 1, 18, 0), datetime.datetime(2023, 1, 2, 12, 0)]

    >>> DatesProvider.from_config(start="2023-01-01 00:00", end="2023-01-02 00:00", frequency=6).as_dict()
    {'start': '2023-01-01T00:00:00', 'end': '2023-01-02T00:00:00', 'frequency': '6h'}

    >>> len(DatesProvider.from_config(start="2023-01-01 00:00", end="2023-01-02 00:00", frequency=12))
    3
    >>> len(DatesProvider.from_config(start="2023-01-01 00:00",
    ...                   end="2023-01-02 00:00",
    ...                   frequency=12,
    ...                   missing=["2023-01-01 12:00"]))
    3
    >>> len(DatesProvider.from_config(start="2023-01-01 00:00",
    ...                   end="2023-01-02 00:00",
    ...                   frequency=12,
    ...                   missing=["2099-01-01 12:00"]))
    3
    """

    def __init__(self, missing: Optional[List[Union[str, datetime.datetime]]] = None) -> None:
        """Initialize the DatesProvider with optional missing dates.

        Parameters
        ----------
        missing : Optional[List[Union[str, datetime.datetime]]]
            List of missing dates.
        """
        if not missing:
            missing = []
        self.missing = list(extend(missing))
        if set(self.missing) - set(self.values):
            diff = set(self.missing) - set(self.values)
            warnings.warn(f"Missing dates {len(diff)=} not in list.")

    @classmethod
    def from_config(cls, **kwargs: Any) -> "DatesProvider":
        """Create a DatesProvider instance from configuration.

        Args:
            **kwargs (Any): Configuration parameters.

        Returns
        -------
        DatesProvider
            An instance of DatesProvider.
        """
        if kwargs.pop("hindcasts", False):
            return HindcastsDates(**kwargs)

        if "values" in kwargs:
            return ValuesDates(**kwargs)

        return StartEndDates(**kwargs)

    def __iter__(self) -> Iterator[datetime.datetime]:
        """Iterate over the dates.

        Yields
        ------
        Iterator[datetime.datetime]
            An iterator of datetime objects.
        """
        yield from self.values

    def __getitem__(self, i: int) -> datetime.datetime:
        """Get a date by index.

        Args:
            i (int): Index of the date.

        Returns
        -------
        datetime.datetime
            The date at the specified index.
        """
        return self.values[i]

    def __len__(self) -> int:
        """Get the number of dates.

        Returns
        -------
        int
            The number of dates.
        """
        return len(self.values)

    @property
    def summary(self) -> str:
        """Get a summary of the date range."""
        return f"📅 {self.values[0]} ... {self.values[-1]}"


class ValuesDates(DatesProvider):
    """Class for handling a list of date values.

    Args:
        values (List[Union[str, datetime.datetime]]): List of date values.
        **kwargs (Any): Additional arguments.
    """

    def __init__(self, values: List[Union[str, datetime.datetime]], **kwargs: Any) -> None:
        """Initialize ValuesDates with a list of values.

        Args:
            values (List[Union[str, datetime.datetime]]): List of date values.
            **kwargs (Any): Additional arguments.
        """
        self.values = sorted([as_datetime(_) for _ in values])
        super().__init__(**kwargs)

    def __repr__(self) -> str:
        """Get a string representation of the ValuesDates instance.

        Returns
        -------
        str
            String representation of the instance.
        """
        return f"{self.__class__.__name__}({self.values[0]}..{self.values[-1]})"

    def as_dict(self) -> Dict[str, Any]:
        """Convert the ValuesDates instance to a dictionary.

        Returns
        -------
        Dict[str, Any]
            Dictionary representation of the instance.
        """
        return {"values": self.values[0]}


class StartEndDates(DatesProvider):
    """Class for generating dates between a start and end date with a specified frequency.

    Args:
        start (Union[str, datetime.datetime]): Start date.
        end (Union[str, datetime.datetime]): End date.
        frequency (Union[int, str]): Frequency of dates.
        **kwargs (Any): Additional arguments.
    """

    def __repr__(self) -> str:
        """Get a string representation of the StartEndDates instance."""
        return f"{self.__class__.__name__}({self.start}..{self.end} every {self.frequency})"

    def __init__(
        self,
        start: Union[str, datetime.datetime],
        end: Union[str, datetime.datetime],
        frequency: Union[int, str] = 1,
        **kwargs: Any,
    ) -> None:
        """Initialize StartEndDates with start, end, and frequency.

        Args:
            start (Union[str, datetime.datetime]): Start date.
            end (Union[str, datetime.datetime]): End date.
            frequency (Union[int, str]): Frequency of dates.
            **kwargs (Any): Additional arguments.
        """
        frequency = frequency_to_timedelta(frequency)
        assert isinstance(frequency, datetime.timedelta), frequency

        def _(x):
            if isinstance(x, str):
                return datetime.datetime.fromisoformat(x)
            return x

        start = _(start)
        end = _(end)

        if isinstance(start, datetime.date) and not isinstance(start, datetime.datetime):
            start = datetime.datetime(start.year, start.month, start.day)

        if isinstance(end, datetime.date) and not isinstance(end, datetime.datetime):
            end = datetime.datetime(end.year, end.month, end.day)

        start = as_datetime(start)
        end = as_datetime(end)

        self.start = start
        self.end = end
        self.frequency = frequency

        missing = kwargs.pop("missing", [])

        self.values = list(DateTimes(start, end, increment=frequency, **kwargs))
        self.kwargs = kwargs

        super().__init__(missing=missing)

    def as_dict(self) -> Dict[str, Any]:
        """Convert the StartEndDates instance to a dictionary.

        Returns
        -------
        Dict[str, Any]
            Dictionary representation of the instance.
        """
        return {
            "start": self.start.isoformat(),
            "end": self.end.isoformat(),
            "frequency": frequency_to_string(self.frequency),
        }.update(self.kwargs)


class Hindcast:
    """Class representing a single hindcast date.

    Args:
        date (datetime.datetime): The date of the hindcast.
        refdate (datetime.datetime): The reference date.
        hdate (datetime.datetime): The hindcast date.
        step (int): The step value.
    """

    def __init__(
        self, date: datetime.datetime, refdate: datetime.datetime, hdate: datetime.datetime, step: int
    ) -> None:
        """Initialize a Hindcast instance.

        Args:
            date (datetime.datetime): The date of the hindcast.
            refdate (datetime.datetime): The reference date.
            hdate (datetime.datetime): The hindcast date.
            step (int): The step value.
        """
        self.date = date
        self.refdate = refdate
        self.hdate = hdate
        self.step = step


class HindcastsDates(DatesProvider):
    """Class for generating hindcast dates over a range of years.

    Args:
        start (Union[str, List[str]]): Start date(s).
        end (Union[str, List[str]]): End date(s).
        steps (List[int]): List of step values.
        years (int): Number of years.
        **kwargs (Any): Additional arguments.
    """

    def __init__(
        self,
        start: Union[str, List[str]],
        end: Union[str, List[str]],
        steps: List[int] = [0],
        years: int = 20,
        **kwargs: Any,
    ) -> None:
        """Initialize HindcastsDates with start, end, steps, and years.

        Args:
            start (Union[str, List[str]]): Start date(s).
            end (Union[str, List[str]]): End date(s).
            steps (List[int]): List of step values.
            years (int): Number of years.
            **kwargs (Any): Additional arguments.
        """
        if not isinstance(start, list):
            start = [start]
            end = [end]

        reference_dates = []
        for s, e in zip(start, end):
            reference_dates.extend(list(DateTimes(s, e, increment=24, **kwargs)))
        # reference_dates = list(DateTimes(start, end, increment=24, **kwargs))
        dates = []

        seen = {}

        for hdate, refdate in HindcastDatesTimes(reference_dates=reference_dates, years=years):
            assert refdate - hdate >= datetime.timedelta(days=365), (refdate - hdate, refdate, hdate)
            for step in steps:

                date = hdate + datetime.timedelta(hours=step)

                if date in seen:
                    raise ValueError(f"Duplicate date {date}={hdate}+{step} for {refdate} and {seen[date]}")

                seen[date] = Hindcast(date, refdate, hdate, step)

                assert refdate - date > datetime.timedelta(days=360), (refdate - date, refdate, date, hdate, step)

                dates.append(date)

        dates = sorted(dates)

        deltas = set()
        for a, b in zip(dates, dates[1:]):
            delta = b - a
            assert isinstance(delta, datetime.timedelta), delta
            deltas.add(delta)

        mindelta_seconds = reduce(gcd, [int(delta.total_seconds()) for delta in deltas])
        mindelta = datetime.timedelta(seconds=mindelta_seconds)
        self.frequency = mindelta
        assert mindelta.total_seconds() > 0, mindelta
        print("🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥", dates[0], dates[-1], mindelta)

        # Use all values between start and end by frequency, and set the ones that are missing
        self.values = []
        missing = []
        date = dates[0]
        last = date
        print("------", date, dates[-1])
        dateset = set(dates)
        while date <= dates[-1]:
            self.values.append(date)
            if date not in dateset:
                missing.append(date)
                seen[date] = seen[last]
            else:
                last = date
            date = date + mindelta

        self.mapping = seen

        print("🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥", self.values[0], self.values[-1], mindelta)
        print("🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥🔥", f"{len(self.values)=} - {len(missing)=}")

        super().__init__(missing=missing)

    def __repr__(self) -> str:
        """Get a string representation of the HindcastsDates instance.

        Returns
        -------
        str
            String representation of the instance.
        """
        return f"{self.__class__.__name__}({self.values[0]}..{self.values[-1]})"

    def as_dict(self) -> Dict[str, Any]:
        """Convert the HindcastsDates instance to a dictionary.

        Returns
        -------
        Dict[str, Any]
            Dictionary representation of the instance.
        """
        return {"hindcasts": self.hindcasts}


if __name__ == "__main__":
    print_dates([datetime.datetime(2023, 1, 1, 0, 0)])
    s = StartEndDates(start="2023-01-01 00:00", end="2023-01-02 00:00", frequency=1)
    print_dates(list(s))



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

import numpy as np
from earthkit.data.core.temporary import temp_file
from earthkit.data.readers.grib.output import new_grib_output

LOG = logging.getLogger(__name__)

CLIP_VARIABLES = (
    "q",
    "cp",
    "lsp",
    "tp",
    "sf",
    "swl4",
    "swl3",
    "swl2",
    "swl1",
)

SKIP = ("class", "stream", "type", "number", "expver", "_leg_number", "anoffset", "time", "date", "step")


def check_compatible(
    f1: Any, f2: Any, centre_field_as_mars: Dict[str, Any], ensemble_field_as_mars: Dict[str, Any]
) -> None:
    """Check if two fields are compatible.

    Parameters
    ----------
    f1 : Any
        The first field.
    f2 : Any
        The second field.
    centre_field_as_mars : Dict[str, Any]
        Metadata of the centre field.
    ensemble_field_as_mars : Dict[str, Any]
        Metadata of the ensemble field.
    """
    assert f1.mars_grid == f2.mars_grid, (f1.mars_grid, f2.mars_grid)
    assert f1.mars_area == f2.mars_area, (f1.mars_area, f2.mars_area)
    assert f1.shape == f2.shape, (f1.shape, f2.shape)

    # Not in *_as_mars
    assert f1.metadata("valid_datetime") == f2.metadata("valid_datetime"), (
        f1.metadata("valid_datetime"),
        f2.metadata("valid_datetime"),
    )

    for k in set(centre_field_as_mars.keys()) | set(ensemble_field_as_mars.keys()):
        if k in SKIP:
            continue
        assert centre_field_as_mars[k] == ensemble_field_as_mars[k], (
            k,
            centre_field_as_mars[k],
            ensemble_field_as_mars[k],
        )


def recentre(
    *,
    members: Any,
    centre: Any,
    clip_variables: Tuple[str, ...] = CLIP_VARIABLES,
    alpha: float = 1.0,
    output: Optional[str] = None,
) -> Any:
    """Recentre ensemble members around the centre field.

    Parameters
    ----------
    members : Any
        The ensemble members.
    centre : Any
        The centre field.
    clip_variables : Tuple[str, ...], optional
        Variables to clip. Defaults to CLIP_VARIABLES.
    alpha : float, optional
        Scaling factor. Defaults to 1.0.
    output : Optional[str], optional
        Output path. Defaults to None.

    Returns
    -------
    Any
        The recentred dataset or output path.
    """
    keys = ["param", "level", "valid_datetime", "date", "time", "step", "number"]

    number_list = members.unique_values("number", progress_bar=False)["number"]
    n_numbers = len(number_list)

    assert None not in number_list

    LOG.info("Ordering fields")
    members = members.order_by(*keys)
    centre = centre.order_by(*keys)
    LOG.info("Done")

    if len(centre) * n_numbers != len(members):
        LOG.error("%s %s %s", len(centre), n_numbers, len(members))
        for f in members:
            LOG.error("Member: %r", f)
        for f in centre:
            LOG.error("centre: %r", f)
        raise ValueError(f"Inconsistent number of fields: {len(centre)} * {n_numbers} != {len(members)}")

    if output is None:
        # prepare output tmp file so we can read it back
        tmp = temp_file()
        path = tmp.path
    else:
        tmp = None
        path = output

    out = new_grib_output(path)

    seen = set()

    for i, centre_field in enumerate(centre):
        param = centre_field.metadata("param")
        centre_field_as_mars = centre_field.metadata(namespace="mars")

        # load the centre field
        centre_np = centre_field.to_numpy()

        # load the ensemble fields and compute the mean
        members_np = np.zeros((n_numbers, *centre_np.shape))

        for j in range(n_numbers):
            ensemble_field = members[i * n_numbers + j]
            ensemble_field_as_mars = ensemble_field.metadata(namespace="mars")
            check_compatible(
                centre_field,
                ensemble_field,
                centre_field_as_mars,
                ensemble_field_as_mars,
            )
            members_np[j] = ensemble_field.to_numpy()

            ensemble_field_as_mars = tuple(sorted(ensemble_field_as_mars.items()))
            assert ensemble_field_as_mars not in seen, ensemble_field_as_mars
            seen.add(ensemble_field_as_mars)

        # cmin=np.amin(centre_np)
        # emin=np.amin(members_np)

        # if cmin < 0 and emin >= 0:
        #     LOG.warning(f"Negative values in {param} cmin={cmin} emin={emin}")
        #     LOG.warning(f"centre: {centre_field_as_mars}")

        mean_np = members_np.mean(axis=0)

        for j in range(n_numbers):
            template = members[i * n_numbers + j]
            e = members_np[j]
            m = mean_np
            c = centre_np

            assert e.shape == c.shape == m.shape, (e.shape, c.shape, m.shape)

            x = c + (e - m) * alpha

            if param in clip_variables:
                # LOG.warning(f"Clipping {param} to be positive")
                x = np.maximum(x, 0)

            assert x.shape == e.shape, (x.shape, e.shape)

            out.write(x, template=template)
            template = None

    assert len(seen) == len(members), (len(seen), len(members))

    out.close()

    if output is not None:
        return path

    from earthkit.data import from_source

    ds = from_source("file", path)

    # save a reference to the tmp file so it is deleted
    # only when the dataset is not used anymore
    ds._tmp = tmp

    assert len(ds) == len(members), (len(ds), len(members))

    return ds
